import struct
import itertools

import envi
import envi.exc as e_exc
import envi.bits as e_bits
import envi.const as e_const
import envi.common as e_common
import envi.memory as e_memory

import visgraph.pathcore as vg_path

import vivisect.exc as v_exc

from vivisect.const import *

import logging
logger = logging.getLogger(__name__)

# Pre-initialize a default stack size
init_stack_size = 0x8000
init_stack_map = b'\xfe' * init_stack_size

def imphook(impname):

    def imptemp(f):
        f.__imphook__ = impname
        return f

    return imptemp

class WorkspaceEmulator:

    taintregs = []

    def __init__(self, vw, **kwargs):
        '''
        Base Emulator Class that other, more platform specific emulators inherit from/mixin.

        Since there's a fair number of knobs that can be turned on the emulator and several
        descendants of the base WorkspaceEmulator, these kwargs apply not only to the base class,
        but most children like the ArmWorkspaceEmulator and any others that live in vivisect/impemu/platarm/

        Current Keyword Arguments:
        * va
            - Type: Integer
            - Default: None
            - Desc: The address to begin emulating from.  emu.setProgramCounter() is called with this value,
                    and in the case of ArmWorkspaceEmulator, appropriate architecture is set.
        * logwrite
            - Type: Boolean
            - Default: False
            - Desc: Enable tracking of all memory writes along the exeuction path. Any results are stored on
                    the path object as a node property at WorkspaceEmulator.curpath
        * logread
            - Type: Boolean
            - Default: False
            - Desc: Enable tracking of all memory writes along the exeuction path. Any results are stored on
                    the path object as a node property at WorkspaceEmulator.curpath
        * safemem
            - Type: Boolean
            - Default: True
            - Desc: Since we're not a real CPU, and since we also do a lot of partial emulation, we can't
                    always be sure we're reading from/writing to a valid memory location. So if safemem is
                    set to True, this enables several addition safety rails in the emulator in terms of
                    where memory can be read from/written to, and where execution flow can be read from
        * funconly
            - Type: Boolean
            - Default: True
            - Desc: When emulating, by default when a call/branch to other function instruction happens,
                    instead of emulating down into the called function, we push the return value onto our
                    emulated stack, and then using the detected calling convention of the subfunction, try
                    and detect where to go. Set this value to False to instead, when a call instruction is
                    hit, emulate down into that subfunction.
                    Please note that since there's not "Stop when you hit X" condition, disabling this
                    can have a massive impact on performance.
        * strictops
            - Type: Boolean
            - Default: True
            - Desc: Due to development time constraints, or plain inability to emulate, not all instructions
                    on every platform we support is emulated. It may be properly decoded, but if it's not
                    not supported, by default when emulating, the emulator will bail out of the paths that
                    contain the unsupported instruction. Set this parameter to false if you want to continue
                    down those paths even when encountering unsupported instructions
                    Please note that since any possible effects the unsupported instruction are not
                    propagated, any possible flags or execution state are most likely misaligned from what
                    a real CPU might experience.
        * loglevel
            - Type: Integer
            - Default: 30 (Corresponds to logging.WARNING)
            - Desc: Because we do partial emulation for several of our heuristic passes like isProbablyCode,
                    but also full emulation for several other analysis passes, there are several logging
                    calls that are erroneous under the heuristic passes and typically we want those silenced
                    under the heuristic passes only.
        * taintbase
            - Type: Integer
            - Default: 0x4156000F
            - Desc: This option affects where we start taint values, which are how we determine what has been
                    written where, what's been called, what values have been edited where, etc.
                    WARNING: Like taintbyte below, while you can modify this value to suit your needs, you
                    should review how the emulator works before modifying this value, as this can directly
                    impact various forms of analysis such as calling convention detection, what values get
                    associated with what API calls, and more.
        * taintbyte
            - Type: A python byte object of length 1
            - Default: b'a'
            - Desc: So when the emulator attempts to read memory, and it's in a location that has yet to be
                    defined (because we're missing the full context of execution inside a real process) or
                    isn't technically considered okay to read from (see probeMemory in envi/memory.py for
                    what we consider to be "okay"), but if we have _safe_mem turned on, we instead return a
                    mocked value that's just the taint bytes, but repeated a number of times equal to the
                    desired read size. Otherwise, we attempt to drop into the base memory object's readMemory
                    to see if there is anything there.
                    WARNING: While this value is configurable, changing this value without knowing what you're
                    doing can result in undesirable effects (such as infinite recursion when trying to repr
                    taint values), so change this value with care.
        * opcache
            - Type: Boolean
            - Default: True
            - Desc: By default, WorkspaceEmulators share the opcode cache storing parsed opcodes for locations
                    in the Vivisect Workspace.
                    This setting disables use of the VivWorkspace's opcode cache.
        '''
        self.vw = vw
        # Set down below in runFunction
        self.funcva = None
        self.emustop = False
        self.usecache = kwargs.get('opcache', True)

        self.hooks = {}
        self.taints = {}
        base = int(kwargs.get('taintbase', 0x4156000F))
        self.taintva = itertools.count(base, 0x2000)
        self.taintoffset = 0x1000
        self.taintmask = 0xffffe000
        self.taintbyte = kwargs.get('taintbyte', b'a')
        self.taintrepr = {}

        self.uninit_use = {}
        self.logwrite = kwargs.get('logwrite', False)
        self.logread = kwargs.get('logread', False)
        self.path = self.newCodePathNode()
        self.curpath = self.path
        self.op = None
        self.emumon = None
        self.psize = self.getPointerSize()

        # Possibly need an "options" API?
        # Should we be forgiving about memory accesses?
        self._safe_mem = kwargs.get("safemem", True)
        # Is this emulator meant to stay in one function scope?
        self._func_only = kwargs.get("funconly", True)
        # Should we bail on emulation if unsupported instruction encountered?
        self.strictops = kwargs.get('strictops', True)
        # So this has nothing to do with the above logwrite/logread
        # But default to being as permissive as the top level logger allows
        self._log_level = kwargs.get('loglevel', logging.WARNING)

        # Map in all the memory associated with the workspace
        for va, size, perms, fname in vw.getMemoryMaps():
            offset, bytes = vw.getByteDef(va)
            self.addMemoryMap(va, perms, fname, bytes)

        for regidx in self.taintregs:
            regval = self.setVivTaint('uninitreg', regidx)
            self.setRegister(regidx, regval)

        for name in dir(self):
            val = getattr(self, name, None)
            if val is None:
                continue

            impname = getattr(val, '__imphook__', None)
            if impname is None:
                continue

            self.hooks[impname] = val

        self.stack_map_mask = None
        self.stack_map_base = None
        self.stack_map_top = None
        self.stack_pointer = None
        self.initStackMemory()

        # set Program Counter if provided
        va = kwargs.get('va')
        if va:
            self.setProgramCounter(va)

    def initStackMemory(self, stacksize=init_stack_size):
        '''
        Setup and initialize stack memory.
        You may call this prior to emulating instructions.
        '''
        if self.stack_map_base is None:
            self.stack_map_mask = e_bits.sign_extend(0xfff00000, 4, self.vw.psize)
            self.stack_map_base = e_bits.sign_extend(0xbfb00000, 4, self.vw.psize)
            self.stack_map_top = self.stack_map_base + stacksize
            self.stack_pointer = self.stack_map_top

            stack_map = init_stack_map
            if stacksize != init_stack_size:
                stack_map = b'\xfe' * stacksize

            # Map in a memory map for the stack

            self.addMemoryMap(self.stack_map_base, 6, "[stack]", stack_map)
            self.setStackCounter(self.stack_pointer)

            # Create some pre-made taints for positive stack indexes
            # NOTE: This is *ugly* for speed....
            taints = [self.setVivTaint('funcstack', i * self.psize) for i in range(20)]
            taintbytes = b''.join([e_bits.buildbytes(taint, self.psize) for taint in taints])

            self.stack_pointer -= len(taintbytes)
            self.setStackCounter(self.stack_pointer)
            self.writeMemory(self.stack_pointer, taintbytes)
        else:
            existing_map_size = self.stack_map_top - self.stack_map_base
            new_map_size = stacksize - existing_map_size
            if new_map_size < 0:
                raise RuntimeError('cannot shrink stack')

            new_map_top = self.stack_map_base
            new_map_base = new_map_top - new_map_size

            stack_map = b''.join([struct.pack('<I', new_map_base+(i*4)) for i in range(new_map_size)])

            self.addMemoryMap(new_map_base, 6, "[stack]", stack_map)
            self.stack_map_base = new_map_base

            # no need to do tainting here, since SP will always be in the
            #   first map

    def stopEmu(self):
        '''
        This is called by monitor to stop emulation
        '''
        self.emustop = True

    def getPathProp(self, key):
        '''
        Retrieve a named value from the current code path context.
        '''
        return vg_path.getNodeProp(self.curpath, key)

    def setPathProp(self, key, value):
        """
        Set a named value which is only relevant for the current code path.
        """
        return vg_path.setNodeProp(self.curpath, key, value)

    def setEmulationMonitor(self, emumon):
        """
        Snap in an emulation monitor. (see EmulationMonitor doc from vivisect.impemu)
        """
        self.emumon = emumon

    def parseOpcode(self, va, arch=envi.ARCH_DEFAULT, skipcache=False):
        if not self.usecache or skipcache:
            return self.__archemu__.parseOpcode(self, va, arch)

        return self.vw.parseOpcode(va, arch=arch)

    def checkCall(self, starteip, endeip, op):
        """
        Check if this was a call, and if so, do the required
        import emulation and such...
        """
        iscall = bool(op.iflags & envi.IF_CALL)
        if iscall:
            # Either way, if it's a call PC goes to next instruction
            thunkReg = self.vw.isFunctionThunkReg(endeip)
            if self._func_only and not thunkReg:
                self.setProgramCounter(starteip+len(op))
            api = self.getCallApi(endeip)
            rtype, rname, convname, callname, funcargs = api
            callconv = self.getCallingConvention(convname)
            if callconv is None:
                return iscall

            argv = callconv.getCallArgs(self, len(funcargs))

            ret = None
            if self.emumon is not None:
                try:
                    ret = self.emumon.apicall(self, op, endeip, api, argv)
                except Exception as e:
                    self.emumon.logAnomaly(self, endeip, "%s.apicall failed: %s" % (self.emumon.__class__.__name__, e))

            hook = self.hooks.get(callname)
            if ret is None and hook:
                hook(self, callconv, api, argv)
            elif self._func_only and not thunkReg:
                if ret is None:
                    ret = self.setVivTaint('apicall', (op, endeip, api, argv))
                retn = self.getProgramCounter()
                callconv.execCallReturn(self, ret, len(funcargs))
                newaddr = self.getProgramCounter()
                # So....sometimes, mostly when we're called into an emulator via isProbablyCode, we don't
                # get the initial function setups, which include a couple pushes onto the stack. Normally,
                # this isn't that much of a problem, but when we hit the last codeblock that includes pops
                # before calling any last few functions, the stack pointer gets throw off by those last few
                # pops, which leads us to say that code path isn't a function since we miss the ret instruction.
                # So we have here a fix for that. Added some rails so we don't always just punch it in
                if self._safe_mem:
                    if not self.vw.isValidPointer(newaddr) and self.isValidPointer(retn):
                        self.setProgramCounter(retn)
            # no else since we'll emulate into the function

        return iscall

    def newCodePathNode(self, parent=None, bva=None):
        '''
        NOTE: Right now, this is only called from the actual branch state which
        needs it.  it must stay that way for now (register context is being copied
        for symbolic emulator...)
        '''
        props = {
            'bva': bva,      # the entry virtual address for this branch
            'valist': [],    # the virtual addresses in this node in order
            'readlog': [],   # a log of all memory reads from this block
            'writelog': [],  # a log of all memory writes from this block
        }
        ret = vg_path.newPathNode(parent=parent, **props)
        return ret

    def getBranchNode(self, node, bva):
        '''
        If a node exists already for the specified branch, return it. Otherwise,
        create a new one and return that...
        '''
        for knode in vg_path.getNodeKids(node):
            if vg_path.getNodeProp(knode, 'bva') == bva:
                return knode
        return self.newCodePathNode(node, bva)

    def checkBranches(self, starteip, endeip, op):
        """
        This routine gets the current branch list for this opcode, adds branch
        entries to the current path, and updates current path as needed
        (returns a list of (va, CodePath) tuples.
        """
        '''
        So I've kinda gone back and forth on this one a bit. On one hand, the emulator
        should be free to iterate over things as it needs. On the other, it's operating with
        imperfect information as to what's a valid path and what's not.
        '''
        vw = self.vw
        ret = []
        paths = set()
        # if we've already hunted this location down, know it's a table, and we've resolved it
        # step carefully so we don't conflate tables together
        xrefs = vw.getXrefsFrom(op.va, rtype=REF_CODE)
        for bva, bflags in op.getBranches(emu=None):
            if bflags & envi.BR_TABLE and vw.getLocation(op.va) and len(xrefs):
                for xrfrom, xrto, xrtype, xrflags in xrefs:
                    # if it's not in a codeblock it's probably something like an import, so we
                    # can ignore it
                    if self.vw.getCodeBlock(xrto) is None:
                        continue
                    if xrto in paths:
                        continue
                    bpath = self.getBranchNode(self.curpath, xrto)
                    ret.append((xrto, bpath))
                    paths.add(xrto)
                return ret

        # FIXME this should actually check for conditional...
        # If there is more than one branch target, we need a new code block
        blist = op.getBranches(emu=self)
        if len(blist) > 1:
            for bva, bflags in blist:
                if bva is None:
                    logger.warning("Unresolved branch even WITH an emulator?")
                    continue
                if bva in paths:
                    continue

                bpath = self.getBranchNode(self.curpath, bva)
                ret.append((bva, bpath))
                paths.add(bva)

        if op.iflags & envi.IF_BRANCH \
                and not op.iflags & envi.IF_COND \
                and len(xrefs):
                    # we've hit a branch that doesn't go anywhere.  probably a switchcase we don't handle here.
                    for xrfr, xrto, xrt, xrflag in xrefs:
                        # skip existing, skip DEREFS
                        if xrto in paths or xrflag & envi.BR_DEREF:
                            continue

                        bpath = self.getBranchNode(self.curpath, xrto)
                        ret.append((xrto, bpath))
                        paths.add(xrto)

        # let's also take into account some of the dynamic branches we may have found
        # like our table pointers
        return ret

    def stepi(self):
        # NOTE: when we step, we *always* want to be stepping over calls
        # (and possibly import emulate them)
        starteip = self.getProgramCounter()

        # parse out an opcode
        op = self.parseOpcode(starteip)
        if self.emumon:
            self.emumon.prehook(self, op, starteip)

        # Execute the opcode
        self.executeOpcode(op)
        vg_path.getNodeProp(self.curpath, 'valist').append(starteip)

        endeip = self.getProgramCounter()

        if self.emumon:
            self.emumon.posthook(self, op, endeip)

        if not self.checkCall(starteip, endeip, op):
            self.checkBranches(starteip, endeip, op)

    def runFunction(self, funcva, stopva=None, maxhit=None, maxloop=None):
        """
        This is a utility function specific to WorkspaceEmulation (and impemu) that
        will emulate, but only inside the given function.  You may specify a stopva
        to return once that location is hit.
        """

        self.funcva = funcva
        # Let the current (should be base also) path know where we are starting
        vg_path.setNodeProp(self.curpath, 'bva', funcva)
        hits = {}
        todo = [(funcva, self.getEmuSnap(), self.path)]
        vw = self.vw  # Save a dereference many many times

        while len(todo):

            va, esnap, self.curpath = todo.pop()

            self.setEmuSnap(esnap)

            self.setProgramCounter(va)

            # Check if we are beyond our loop max...
            if maxloop is not None:
                lcount = vg_path.getPathLoopCount(self.curpath, 'bva', va)
                if lcount > maxloop:
                    continue

            while True:

                starteip = self.getProgramCounter()

                if not vw.isValidPointer(starteip):
                    break

                if starteip == stopva:
                    return

                # Check straight hit count...
                if maxhit is not None:
                    h = hits.get(starteip, 0)
                    h += 1
                    if h > maxhit:
                        break
                    hits[starteip] = h

                # If we ran out of path (branches that went
                # somewhere that we couldn't follow)?
                if self.curpath is None:
                    break

                try:
                    # FIXME unify with stepi code...
                    op = self.parseOpcode(starteip)
                    self.op = op
                    if self.emumon:
                        try:
                            self.emumon.prehook(self, op, starteip)
                        except v_exc.BadOpBytes as e:
                            logger.debug(str(e))
                            break
                        except v_exc.BadOutInstruction:
                            pass
                        except Exception as e:
                            logger.log(self._log_level, "Emulator prehook failed on fva: 0x%x, opva: 0x%x, op: %s, err: %s", funcva, starteip, str(op), str(e))

                        if self.emustop:
                            return
                    # Execute the opcode
                    self.executeOpcode(op)
                    vg_path.getNodeProp(self.curpath, 'valist').append(starteip)

                    endeip = self.getProgramCounter()

                    if self.emumon:
                        try:
                            self.emumon.posthook(self, op, endeip)
                        except v_exc.BadOpBytes as e:
                            logger.debug(str(e))
                            break
                        except v_exc.BadOutInstruction:
                            pass
                        except Exception as e:
                            logger.log(self._log_level, "funcva: 0x%x opva: 0x%x:  %r   (%r) (in emumon posthook)", funcva, starteip, op, e)

                        if self.emustop:
                            return

                    iscall = self.checkCall(starteip, endeip, op)
                    if self.emustop:
                        return

                    # If it wasn't a call, check for branches, if so, add them to
                    # the todo list and go around again...
                    if not iscall:
                        blist = self.checkBranches(starteip, endeip, op)
                        if len(blist):
                            # pc in the snap will be wrong, but over-ridden at restore
                            esnap = self.getEmuSnap()
                            for bva, bpath in blist:
                                todo.append((bva, esnap, bpath))
                            break

                    # If we enounter a procedure exit, it doesn't
                    # matter what EIP is, we're done here.
                    if op.iflags & envi.IF_RET:
                        vg_path.setNodeProp(self.curpath, 'cleanret', True)
                        break

                    # TODO: hook things like error(...) when they have a param that indicates to
                    # exit. Might be a bit hairy since we'll possibly have to fix up codeblocks
                    # Make sure we can at least get past the first instruction in certain functions
                    if self.vw.isNoReturnVa(op.va) and op.va != funcva:
                        vg_path.setNodeProp(self.curpath, 'cleanret', False)
                        break

                except envi.BadOpcode:
                    break
                except envi.UnsupportedInstruction as e:
                    if self.strictops:
                        logger.debug('runFunction failed: unsupported instruction: 0x%08x %s', e.op.va, e.op.mnem)
                        break
                    else:
                        logger.debug('runFunction continuing after unsupported instruction: 0x%08x %s', e.op.va, e.op.mnem)
                        self.setProgramCounter(e.op.va + e.op.size)
                except v_exc.BadOutInstruction:
                    break
                except Exception as e:
                    if self.emumon is not None and not isinstance(e, e_exc.BreakpointHit):
                        self.emumon.logAnomaly(self, starteip, str(e))

                    break  # If we exc during execution, this branch is dead.

    def getCallApi(self, va):
        '''
        Retrieve an API definition from either the vivisect workspace
        ( if the call target is a function within the workspace ) or
        the impapi definition subsystem ( if the call target is a known
        import definition )
        '''
        vw = self.vw
        ret = None

        if vw.isFunction(va):
            ret = vw.getFunctionApi(va)
            if ret is not None:
                return ret

        else:

            taint = self.getVivTaint(va)
            if taint:
                tva,ttype,tinfo = taint

                if ttype == 'import':
                    lva,lsize,ltype,linfo = tinfo
                    ret = vw.getImpApi( linfo )

                elif ttype == 'dynfunc':
                    libname,funcname = tinfo
                    ret = vw.getImpApi('%s.%s' % (libname,funcname))

                if ret:
                    return ret

        defcall = vw.getMeta("DefaultCall")
        return ('int', None, defcall, 'UnknownApi', () )

    def nextVivTaint(self):
        # One page into the new taint range
        return next(self.taintva) + self.taintoffset

    def setVivTaint(self, typename, taint):
        '''
        Set a taint in the emulator.  Returns the new value for
        the created taint.
        '''
        va = self.nextVivTaint()
        self.taints[ va & self.taintmask ] = (va,typename,taint)
        return va

    def getVivTaint(self, va):
        '''
        Retrieve a previously registered taint ( this will automagically
        mask values down and allow you to retrieve "near taint" values.)
        '''
        return self.taints.get( va & self.taintmask )

    def reprVivTaint(self, taint):
        '''
        For the base "known" taint types, return a humon readable string
        to represent the value of the taint.
        '''

        va, ttype, tinfo = taint

        if ttype == 'uninitreg':
            trepr = self.getRegisterName(tinfo)
        elif ttype == 'import':
            lva,lsize,ltype,linfo = tinfo
            trepr = linfo
        elif ttype == 'dynlib':
            libname = tinfo
            trepr = libname
        elif ttype == 'dynfunc':
            libname,funcname = tinfo
            trepr = '%s.%s' % (libname,funcname)
        elif ttype == 'funcstack':
            stackoff = tinfo
            if self.funcva:
                flocal = self.vw.getFunctionLocal(self.funcva, stackoff)
                if flocal is not None:
                    typename,argname = flocal
                    return argname
            o = '+'
            if stackoff < 0:
                o = '-'
            trepr = 'sp%s%d' % (o, abs(stackoff))
        elif ttype == 'apicall':
            op, pc, api, argv = tinfo
            if op.va in self.taintrepr:
                return '<0x%.8x>' % op.va
            rettype, retname, callconv, callname, callargs = api
            callstr = self.reprVivValue(pc)
            argsstr = ','.join([self.reprVivValue(x) for x in argv])
            trepr = '%s(%s)' % (callstr, argsstr)
            self.taintrepr[op.va] = trepr
        else:
            trepr = 'taint: 0x%.8x %s %r' % (va, ttype, tinfo)

        return trepr

    def reprVivValue(self, val):
        '''
        Return a humon readable string which is the best description for
        the given value ( given knowledge of the workspace, emu,
        and taint subsystems ).
        '''
        if self.vw.isFunction(val):
            thunk = self.vw.getFunctionMeta(val, 'Thunk')
            if thunk:
                return thunk

        vivname = self.vw.getName(val)
        if vivname:
            return vivname

        taint = self.getVivTaint(val)
        if taint:
            va, ttype, tinfo = taint
            if ttype == 'apicall':
                op, pc, api, argv = tinfo
                rettype, retname, callconv, callname, callargs = api
                if val not in argv:
                    return self.reprVivTaint(taint)

            else:
                return self.reprVivTaint(taint)


        stackoff = self.getStackOffset(val)
        if stackoff is not None:
            funclocal = self.vw.getFunctionLocal(self.funcva, stackoff)
            if funclocal is not None:
                typename, varname = funclocal
                return varname

        if val < 4096:
            return str(val)

        return '0x%.8x' % val

    def _useVirtAddr(self, va):
        taint = self.getVivTaint(va)
        if taint is None:
            return

        tva,ttype,tinfo = taint

        if ttype == 'uninitreg':
            self.logUninitRegUse(tinfo)

    def writeMemory(self, va, bytes):
        """
        Try to write the bytes to the memory object, otherwise, dont'
        complain...
        """
        if self.logwrite:
            wlog = vg_path.getNodeProp(self.curpath, 'writelog')
            wlog.append((self.getProgramCounter(), va, bytes))

        self._useVirtAddr(va)

        # It's totally ok to write to invalid memory during the
        # emulation pass (as long as safe_mem is true...)
        probeok = self.probeMemory(va, len(bytes), e_const.MM_WRITE)
        if self._safe_mem and not probeok:
            return

        return e_memory.MemoryObject.writeMemory(self, va, bytes)

    def logUninitRegUse(self, regid):
        self.uninit_use[regid] = True

    def getUninitRegUse(self):
        return list(self.uninit_use.keys())

    def readMemory(self, va, size):
        if self.logread:
            rlog = vg_path.getNodeProp(self.curpath, 'readlog')
            rlog.append((self.getProgramCounter(), va, size))

        # If they read an import entry, start a taint...
        loc = self.vw.getLocation(va)
        if loc is not None:
            lva, lsize, ltype, ltinfo = loc
            if ltype == LOC_IMPORT and lsize == size:  # They just read an import.
                ret = self.setVivTaint('import', loc)
                return e_bits.buildbytes(ret, lsize)

        self._useVirtAddr(va)

        # Read from the emulator's pages if we haven't resolved it yet
        probeok = self.probeMemory(va, size, e_const.MM_READ)
        if self._safe_mem and not probeok:
            return self.taintbyte * size

        return e_memory.MemoryObject.readMemory(self, va, size)

    # Some APIs for telling if pointers are in runtime memory regions

    def isUninitStack(self, val):
        """
        If val is a numerical value in the same memory page
        as the un-initialized stack values return True
        """
        # NOTE: If uninit_stack_byte changes, so must this!
        if (val & 0xfffff000) == 0xfefef000:
            return True
        return False

    def isStackPointer(self, va):
        return (va & self.stack_map_mask) == self.stack_map_base

    def getStackOffset(self, va):
        if (va & self.stack_map_mask) == self.stack_map_base:
            return va - self.stack_pointer




"""
The initial arm module.
"""

import envi

from envi.archs.arm.regs import *
from envi.archs.arm.disasm import *

class ArmModule(envi.ArchitectureModule):

    def __init__(self, name='arm'):
        import envi.archs.thumb16.disasm as eatd
        # these are required for setEndian() which is called from ArchitectureModule.__init__()
        self._arch_dis = ArmDisasm()
        self._arch_dis.setArchMask(name)
        self._arch_thumb_dis = eatd.ThumbDisasm()

        envi.ArchitectureModule.__init__(self, name, maxinst=4)
        self._arch_reg = self.archGetRegCtx()

        # pre-generate this list
        self.badoplist = [self.archParseOpcode(badop, 0, 0) for badop in self._arch_badopbytes]
        self.badoplist.extend([self.archParseOpcode(badop, 0, 1) for badop in self._arch_badopbytes])

    def initRegGroups(self):
        envi.ArchitectureModule.initRegGroups(self)
        self._regGrps.update({'general': arm_regs})

    def archGetRegCtx(self):
        return ArmRegisterContext()

    def archGetBreakInstr(self):
        raise Exception ("weird... what are you trying to do here?  ARM has a complex breakpoint instruction")

    def archGetNopInstr(self):
        return (b'\x00\x00\x60\xe3', b'\xe3\x60\x00\x00')[self._endian]   #FIXME: this is only ARM mode.  this arch mod should cover both.  the ENVI architecture doesn't support this model yet.

    def archGetBadOps(self):
        return self.badoplist

    def getPointerSize(self):
        return 4

    def pointerString(self, va):
        return "0x%.8x" % va

    def archParseOpcode(self, bytes, offset=0, va=0, extra=None):
        """
        Parse a sequence of bytes out into an envi.Opcode instance.
        """
        if va & 3:
            offset &= -2
            va &= -2

            return self._arch_thumb_dis.disasm(bytes, offset, va)

        return self._arch_dis.disasm(bytes, offset, va)

    def getEmulator(self):
        return ArmEmulator()

    def setEndian(self, endian):
        self._endian = endian
        self._arch_dis.setEndian(endian)
        self._arch_thumb_dis.setEndian(endian)

    def archModifyFuncAddr(self, va, info):
        if va & 1:
            return va & -2, {'arch' : envi.ARCH_THUMB}
        return va, info

    def archModifyXrefAddr(self, tova, reftype, rflags=0):
        # if rflags is non-zero, it indicates that the target is Code
        if tova & 3 and rflags:
            return tova & -2, reftype, rflags
        return tova, reftype, rflags

    def archGetPointerAlignment(self):
        return 4


class ThumbModule(envi.ArchitectureModule):
    '''
    This architecture module will *not* shift to ARM mode.  Evar.
    '''

    def __init__(self, name='thumb'):
        import envi.archs.thumb16.disasm as eatd
        # this is required for setEndian() which is called from ArchitectureModule.__init__()
        self._arch_dis = eatd.ThumbDisasm(doModeSwitch=False)

        envi.ArchitectureModule.__init__(self, name, maxinst=4)
        self._arch_reg = self.archGetRegCtx()
        #armVersion mask should be set here if needed

        # pre-generating bad-ops list
        self.badoplist = [ self.archParseOpcode(badop,0,0) for badop in self._arch_badopbytes ]
        self.badoplist.extend([ self.archParseOpcode(badop,0,1) for badop in self._arch_badopbytes ])

    def initRegGroups(self):
        envi.ArchitectureModule.initRegGroups(self)
        self._regGrps.update({'general': arm_regs})

    def archGetRegCtx(self):
        return ArmRegisterContext()

    def archGetBreakInstr(self):
        raise Exception ("weird... what are you trying to do here?  ARM has a complex breakpoint instruction")

    def archGetNopInstr(self):
        return (b'\xc0\x46', b'\x46\xc0')[self._endian]

    def archGetBadOps(self):
        return self.badoplist

    def getPointerSize(self):
        return 4

    def pointerString(self, va):
        return "0x%.8x" % va

    def archParseOpcode(self, bytes, offset=0, va=0, extra=None):
        """
        Parse a sequence of bytes out into an envi.Opcode instance.
        """
        va &= -2
        return self._arch_dis.disasm(bytes, offset, va)

    def getEmulator(self):
        emu = ArmEmulator()
        emu.setThumbMode()
        return emu

    def setEndian(self, endian):
        self._endian = endian
        self._arch_dis.setEndian(endian)

    def archModifyFuncAddr(self, va, info):
        if va & 1:
            return va & -2, {'arch' : envi.ARCH_THUMB}
        return va, info

    def archModifyXrefAddr(self, tova, reftype, rflags):
        if tova & 1:
            return tova & -2, reftype, rflags
        return tova, reftype, rflags

    def archGetPointerAlignment(self):
        return 4


from envi.archs.arm.emu import *



import envi
import platform
import unittest

def skip(*skips):
    arch = envi.getCurrentArch()
    plat = platform.system().lower()

    cur = set([arch,plat])
    skips = set(skips)

    def skipfunc(f):
        has = cur & skips
        if not has:
            return f

        def doskip(*args, **kwargs):
            raise unittest.SkipTest('Skipped For: %s' % (repr(has),))

        return doskip

    return skipfunc




from PyQt5.QtWidgets import *

import vqt.main as vq_main
import vqt.tree as vq_tree

import envi.threads as e_threads
import cobra.remoteapp as c_remoteapp
import vivisect.remote.server as viv_server
import vivisect.qt.funcgraph as v_q_funcgraph

from vqt.basics import *


class WorkspaceListModel(vq_tree.VQTreeModel):
    columns = ('Name',)


class WorkspaceListView(vq_tree.VQTreeView):
    def __init__(self, workspaces, parent=None):
        vq_tree.VQTreeView.__init__(self, parent=parent)
        model = WorkspaceListModel(parent=self)
        self.setModel(model)
        for wsname in workspaces:
            model.append((wsname,))


class BaseServerDialog(QDialog):

    def __init__(self, workspaces, parent=None):
        QDialog.__init__(self, parent=parent)
        self.setWindowTitle('Select a workspace...')

        self.wsname = None
        self.wslist = WorkspaceListView(workspaces, parent=self)

        self.buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        self.buttons.accepted.connect(self.accept)
        self.buttons.rejected.connect(self.reject)

        layout = VBox()
        layout.addWidget(self.wslist)
        layout.addWidget(self.buttons)
        self.setLayout(layout)

        self.wslist.doubleClicked.connect( self.workspaceActivated )

    def getWorkspaceName(self):
        self.exec_()
        return self.wsname

    def workspaceActivated(self, idx):
        self.accept()

    def accept(self):
        for idx in self.wslist.selectedIndexes():
            row = idx.internalPointer()
            if row:
                self.wsname = row.rowdata[0]
                break

        return QDialog.accept(self)


class VivServerDialog(QDialog):
    def __init__(self, vw, parent=None):
        QDialog.__init__(self, parent=parent)
        self.vw = vw
        try:
            server = vw.config.remote.server
        except AttributeError:
            server = "visi.kenshoto.com"

        self.wsserver = QLineEdit(server, parent=self)
        self.setdef = QCheckBox(parent=self)

        self.buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        self.buttons.accepted.connect(self.accept)
        self.buttons.rejected.connect(self.reject)

        serverlayout = QHBoxLayout()
        serverlayout.addWidget(self.wsserver)
        serverlayout.addWidget(QLabel('Make Default:'))
        serverlayout.addWidget(self.setdef)

        layout = QFormLayout()
        layout.addRow('Workspace Server', serverlayout)
        layout.addWidget(self.buttons)
        self.setLayout(layout)

    def accept(self, *args, **kwargs):
        QDialog.accept(self, *args, **kwargs)
        if self.setdef.isChecked():
            cfg = self.vw.config.getSubConfig("remote")
            cfg['server'] = str(self.wsserver.text())
            self.vw.config.saveConfigFile()

class VivSaveServerDialog(VivServerDialog):

    def __init__(self, vw, parent=None):
        VivServerDialog.__init__(self, vw, parent=parent)

        self.wsname = QLineEdit(vw.getMeta('StorageName', ''), parent=self)
        wslabel = QLabel("Storage Path:")

        bot_box = QWidget(parent=self)
        hbox = QHBoxLayout(bot_box)
        hbox.setContentsMargins(2, 2, 2, 2)
        hbox.setSpacing(4)
        hbox.addWidget(wslabel)
        hbox.addWidget(self.wsname)

        layout = self.layout()
        layout.insertRow(1, bot_box)

        self.setWindowTitle('Save to Workspace Server...')

    def getNameAndServer(self):
        if not self.exec_():
            return (None, None)
        wsname = str(self.wsname.text())
        wsserver = str(self.wsserver.text())
        return (wsname, wsserver)


class VivConnectServerDialog(VivServerDialog):
    def __init__(self, vw, parent=None):
        VivServerDialog.__init__(self, vw, parent=parent)
        self.setWindowTitle('Workspace Server...')

    def getServer(self):
        if not self.exec_():
            return None
        wsserver = str(self.wsserver.text())
        return wsserver


def openServerAndWorkspace(vw, parent=None):
    dia = VivConnectServerDialog(vw, parent=parent)
    host = dia.getServer()
    if host is None:
        return
    port = viv_server.viv_port
    if ':' in host:
        host, port = host.split(':')
        port = int(port)

    connServerAndWorkspace(vw, str(host), port, parent=parent)


def connServerAndWorkspace(vw, host, port=viv_server.viv_port, parent=None):
    # NOTE: do *not* touch parent (or qt) in here!
    try:
        server = viv_server.connectToServer(host, port=port)
        wslist = server.listWorkspaces()
        selectServerWorkspace(vw, server, wslist, parent=parent)
    except Exception as e:
        vw.vprint('Server Error: %s' % e)
        return

def selectServerWorkspace(vw, server, workspaces, parent=None):
    dia = BaseServerDialog(workspaces, parent=parent)
    workspace = dia.getWorkspaceName()
    if workspace is None:
        return

    loadServerWorkspace(vw, server, workspace)

@vq_main.workthread
def loadServerWorkspace(oldvw, server, workspace):
    oldvw.vprint('Loading Workspace: %s' % workspace)
    vw = viv_server.getServerWorkspace(server, workspace)
    # reset the FunctionGraph Counter
    v_q_funcgraph.reset()

    import vivisect.qt.main as viv_q_main
    viv_q_main.runqt(vw, closeme=oldvw.getVivGui())

def saveToServer(vw, parent=None):
    dia = VivSaveServerDialog(vw, parent=parent)
    wsname, wsserver = dia.getNameAndServer()

    vw.vprint('Saving to Workspace Server: %s (%s)' % (wsserver,wsname))
    if wsserver is None:
        return
    port = viv_server.viv_port
    if ':' in wsserver:
        wsserver, port = wsserver.split(':')
        port = int(port)

    sendServerWorkspace(vw, wsname, wsserver, port)

@e_threads.firethread
def sendServerWorkspace(vw, wsname, wsserver, port=viv_server.viv_port):
    try:
        events = vw.exportWorkspace()
        server = viv_server.connectToServer(wsserver, port)
        server.addNewWorkspace(wsname, events)

    except Exception as e:
        vw.vprint('Workspace Server Error: %s' % e)
        import traceback;
        traceback.print_exc()
        return

    vw.setMeta('WorkspaceServer', wsserver)

def openSharedWorkspace(vw, parent=None):
    '''
    Open a workspace shared by a vivisect peer.
    '''
    hostport, ok = QInputDialog.getText(parent, 'Shared Workspace...', 'host:port')
    if not ok:
        return

    uri = 'cobra://%s/vivisect.remote.client?msgpack=1' % hostport
    c_remoteapp.execRemoteApp(uri)





# The opcode tables were taken from Mammon_'s Guide to Writing Disassemblers in Perl, You Morons!"
# and the bastard project. http://www.eccentrix.com/members/mammon/

INSTR_PREFIX = 0xF0000000
ADDRMETH_MASK = 0x00FF0000
ADDRMETH_A = 0x00010000    # Direct address with segment prefix
ADDRMETH_B = 0x00020000    # VEX.vvvv field selects general purpose register
ADDRMETH_C = 0x00030000    # MODRM reg field defines control register
ADDRMETH_D = 0x00040000    # MODRM reg field defines debug register
ADDRMETH_E = 0x00050000    # MODRM byte defines reg/memory address
ADDRMETH_F = 0x00060000    # EFLAGS/RFLAGS register
ADDRMETH_G = 0x00070000    # MODRM byte defines general-purpose reg
ADDRMETH_H = 0x00080000    # VEX.vvvv field selects 128bit XMM or 256bit YMM register
ADDRMETH_I = 0x00090000    # Immediate data follows
ADDRMETH_J = 0x000A0000    # Immediate value is relative to EIP
ADDRMETH_L = 0x000B0000    # An immediate value follows, but bits[7:4] signifies an xmm register
ADDRMETH_M = 0x000C0000    # MODRM mod field can refer only to memory
ADDRMETH_N = 0x000D0000    # R/M field of MODRM selects a packed-quadword, MMX register
ADDRMETH_O = 0x000E0000    # Displacement follows (without modrm/sib)
ADDRMETH_P = 0x000F0000    # MODRM reg field defines MMX register
ADDRMETH_Q = 0x00100000    # MODRM defines MMX register or memory
ADDRMETH_R = 0x00110000    # MODRM mod field can only refer to register
ADDRMETH_S = 0x00120000    # MODRM reg field defines segment register
ADDRMETH_U = 0x00130000    # MODRM R/M field defines XMM register
ADDRMETH_V = 0x00140000    # MODRM reg field defines XMM register
ADDRMETH_W = 0x00150000    # MODRM defines XMM register or memory
ADDRMETH_X = 0x00160000    # Memory addressed by DS:rSI
ADDRMETH_Y = 0x00170000    # Memory addressd by ES:rDI
ADDRMETH_VEXH = 0x00180000  # Maybe Ignore the VEX.vvvv field based on what the ModRM bytes are
ADDRMETH_LAST = ADDRMETH_VEXH

ADDRMETH_VEXSKIP = 0x00800000  # This operand should be skipped if we're not in VEX mode

OPTYPE_a = 0x01000000     # 2/4   two one-word operands in memory or two double-word operands in memory (operand-size attribute)   
OPTYPE_b = 0x02000000     # 1     always 1 byte
OPTYPE_c = 0x03000000     # 1/2   byte or word, depending on operand
OPTYPE_d = 0x04000000     # 4     double-word
OPTYPE_ds = 0x04000000     # 4     double-word
OPTYPE_dq = 0x05000000     # 16    double quad-word
OPTYPE_p = 0x06000000     # 4/6   32-bit or 48-bit pointer
OPTYPE_pi = 0x07000000     # 8     quadword MMX register
OPTYPE_ps = 0x08000000     # 16    128-bit single-precision float (packed?)
OPTYPE_pd = 0x08000000     # ??    should be a double-precision float?
OPTYPE_q = 0x09000000     # 8     quad-word
OPTYPE_qp = 0x09000000     # 8     quad-word
OPTYPE_qq = 0x0A000000     # 8     quad-word
OPTYPE_s = 0x0B000000     # 6     6-byte pseudo descriptor
OPTYPE_ss = 0x0C000000     # ??    Scalar of 128-bit single-precision float
OPTYPE_si = 0x0D000000     # 4     Doubleword integer register
OPTYPE_sd = 0x0E000000     # Scalar double precision float
OPTYPE_v = 0x0F000000     # 2/4   word or double-word, depending on operand
OPTYPE_w = 0x10000000     # 2     always word
OPTYPE_x = 0x11000000     # 2     double-quadword or quad-quadword
OPTYPE_y = 0x12000000     # 4/8   dword or qword
OPTYPE_z = 0x13000000     # 2/4   is this OPTYPE_z?  word for 16-bit operand size or doubleword for 32 or 64-bit operand-size

OPTYPE_fs = 0x14000000
OPTYPE_fd = 0x15000000
OPTYPE_fe = 0x16000000
OPTYPE_fb = 0x17000000
OPTYPE_fv = 0x18000000

# FIXME this should probably be a list rather than a dictionary

OPERSIZE = {
    0: (2, 4, 8),           # We will only end up here on regs embedded in opcodes
    OPTYPE_a: (4, 8, 8),
    OPTYPE_b: (1, 1, 1),
    OPTYPE_c: (1, 2, 2),           # 1/2   byte or word, depending on operand
    OPTYPE_d: (4, 4, 4),           # 4     double-word
    OPTYPE_dq: (16, 16, 16),        # 16    double quad-word
    OPTYPE_p: (4, 6, 6),           # 4/6   32-bit or 48-bit pointer
    OPTYPE_pi: (8, 8, 8),           # 8     quadword MMX register
    OPTYPE_ps: (16, 16, 16),        # 16    128-bit single-precision float
    OPTYPE_pd: (16, 16, 16),        # ??    should be a double-precision float?
    OPTYPE_q: (8, 8, 8),           # 8     quad-word
    OPTYPE_qq: (32, 32, 32),        # 32    quad-quad-word
    OPTYPE_s: (6, 10, 10),         # 6     6-byte pseudo descriptor
    OPTYPE_ss: (16, 16, 16),        # ??    Scalar of 128-bit single-precision float
    OPTYPE_si: (4, 4, 4),           # 4     Doubleword integer register
    OPTYPE_sd: (16, 16, 16),        # ???   Scalar of 128-bit double-precision float
    OPTYPE_v: (2, 4, 8),           # 2/4   word or double-word, depending on operand
    OPTYPE_w: (2, 2, 2),           # 2     always word
    OPTYPE_x: (16, 16, 32),        # 16/32 double-quadword or quad-quadword
    OPTYPE_y: (4, 4, 8),           # 4/8   dword or qword in 64-bit mode
    OPTYPE_z: (2, 4, 4),           # word for 16-bit operand size or doubleword for 32 or 64-bit operand-size
    # Floating point crazyness FIXME these are mostly wrong
    OPTYPE_fs: (4, 4, 4),
    OPTYPE_fd: (8, 8, 8),
    OPTYPE_fe: (10, 10, 10),
    OPTYPE_fb: (10, 10, 10),
    OPTYPE_fv: (14, 14, 28),
}

INS_NOPREF = 0x10000  # This instruction diallows prefixes, and if it does, it's a different insttruction
INS_VEXREQ = 0x20000  # This instructions requires VEX
INS_VEXNOPREF = 0x40000  # This instruction doesn't get the "v" prefix common to VEX instructions

INS_EXEC = 0x1000
INS_ARITH = 0x2000
INS_LOGIC = 0x3000
INS_STACK = 0x4000
INS_COND = 0x5000
INS_LOAD = 0x6000
INS_ARRAY = 0x7000
INS_BIT = 0x8000
INS_FLAG = 0x9000
INS_FPU = 0xA000
INS_TRAPS = 0xD000
INS_SYSTEM = 0xE000
INS_OTHER = 0xF000

INS_BRANCH = INS_EXEC | 0x01
INS_BRANCHCC = INS_EXEC | 0x02
INS_CALL = INS_EXEC | 0x03
INS_CALLCC = INS_EXEC | 0x04
INS_RET = INS_EXEC | 0x05
INS_LOOP = INS_EXEC | 0x06
INS_ADD = INS_ARITH | 0x01

INS_SUB = INS_ARITH | 0x02
INS_MUL = INS_ARITH | 0x03
INS_DIV = INS_ARITH | 0x04
INS_INC = INS_ARITH | 0x05
INS_DEC = INS_ARITH | 0x06
INS_SHL = INS_ARITH | 0x07
INS_SHR = INS_ARITH | 0x08
INS_ROL = INS_ARITH | 0x09
INS_ROR = INS_ARITH | 0x0A
INS_ABS = INS_ARITH | 0x0B

INS_AND = INS_LOGIC | 0x01
INS_OR = INS_LOGIC | 0x02
INS_XOR = INS_LOGIC | 0x03
INS_NOT = INS_LOGIC | 0x04
INS_NEG = INS_LOGIC | 0x05

INS_PUSH =        INS_STACK | 0x01
INS_POP =         INS_STACK | 0x02
INS_PUSHREGS =    INS_STACK | 0x03
INS_POPREGS =     INS_STACK | 0x04
INS_PUSHFLAGS =   INS_STACK | 0x05
INS_POPFLAGS =    INS_STACK | 0x06
INS_ENTER =       INS_STACK | 0x07
INS_LEAVE =       INS_STACK | 0x08

INS_TEST  =              INS_COND | 0x01
INS_CMP   =      INS_COND | 0x02
INS_MOV    =     INS_LOAD | 0x01

INS_MOVCC  =             INS_LOAD | 0x02
INS_XCHG   =             INS_LOAD | 0x03
INS_XCHGCC =     INS_LOAD | 0x04
INS_LEA    =     INS_LOAD | 0x05

INS_STRCMP  = INS_ARRAY | 0x01
INS_STRLOAD = INS_ARRAY | 0x02
INS_STRMOV  = INS_ARRAY | 0x03
INS_STRSTOR = INS_ARRAY | 0x04
INS_XLAT    = INS_ARRAY | 0x05

INS_BITTEST =  INS_BIT | 0x01
INS_BITSET  =  INS_BIT | 0x02
INS_BITCLR  =  INS_BIT | 0x03

INS_CLEARCF  = INS_FLAG | 0x01
INS_CLEARZF  = INS_FLAG | 0x02
INS_CLEAROF  = INS_FLAG | 0x03
INS_CLEARDF  = INS_FLAG | 0x04
INS_CLEARSF  = INS_FLAG | 0x05
INS_CLEARPF  = INS_FLAG | 0x06
INS_SETCF    = INS_FLAG | 0x07
INS_SETZF    = INS_FLAG | 0x08
INS_SETOF    = INS_FLAG | 0x09
INS_SETDF    = INS_FLAG | 0x0A
INS_SETSF    = INS_FLAG | 0x0B
INS_SETPF    = INS_FLAG | 0x0C
INS_CLEARAF  = INS_FLAG | 0x0D
INS_SETAF    = INS_FLAG | 0x0E

INS_TOGCF    = INS_FLAG | 0x10  # /* toggle */
INS_TOGZF    = INS_FLAG | 0x20
INS_TOGOF    = INS_FLAG | 0x30
INS_TOGDF    = INS_FLAG | 0x40
INS_TOGSF    = INS_FLAG | 0x50
INS_TOGPF    = INS_FLAG | 0x60

INS_TRAP = INS_TRAPS | 0x01         # generate trap
INS_TRAPCC = INS_TRAPS | 0x02       # conditional trap gen
INS_TRET = INS_TRAPS | 0x03         # return from trap
INS_BOUNDS = INS_TRAPS | 0x04       # gen bounds trap
INS_DEBUG = INS_TRAPS | 0x05        # gen breakpoint trap
INS_TRACE = INS_TRAPS | 0x06        # gen single step trap
INS_INVALIDOP = INS_TRAPS | 0x07    # gen invalid instruction
INS_OFLOW = INS_TRAPS | 0x08       # gen overflow trap

#/* INS_SYSTEM */
INS_HALT    = INS_SYSTEM | 0x01 # halt machine
INS_IN      = INS_SYSTEM | 0x02 # input from port
INS_OUT     = INS_SYSTEM | 0x03 # output to port
INS_CPUID   = INS_SYSTEM | 0x04 # iden
INS_RDTSC   = INS_SYSTEM | 0x05 # read timer
INS_RDTSCP  = INS_SYSTEM | 0x06 # read timer
INS_SYSCALL = INS_SYSTEM | 0x07 # system call
INS_INT     = INS_SYSTEM | 0x08 # system call
INS_SYSENTER= INS_SYSTEM | 0x09 # system call
INS_SYSEXIT = INS_SYSTEM | 0x0a # system call

INS_NOP     = INS_OTHER | 0x01
INS_BCDCONV = INS_OTHER | 0x02  # convert to/from BCD
INS_SZCONV  = INS_OTHER | 0x03  # convert size of operand
INS_CRYPT   = INS_OTHER | 0x04  # AES-NI instruction support

# string instructions that support REP prefix
REP_OPCODES = (
    INS_IN,       # INS
    INS_OUT,      # OUTS
    INS_STRMOV,   # MOVS
    INS_STRLOAD,  # LODS
    INS_STRSTOR,  # STOS
    INS_STRCMP    # CMPS, SCAS
    )

OP_R = 0x001
OP_W = 0x002
OP_X = 0x004
OP_64AUTO = 0x008  # operand is in 64bit mode with amd64!
# So these this exists is because in the opcode mappings intel puts out, they very 
# *specifically* call out things like pmovsx* using U/M for their operand mappings,
# but *not* W. The reason for this being there
# is a size difference between the U and M portions, whereas W uses a uniform size for both
OP_MEM_B = 0x010  # force only *memory* to be 8 bit.
OP_MEM_W = 0x020  # force only *memory* to be 16 bit.
OP_MEM_D = 0x030  # force only *memory* to be 32 bit.
OP_MEM_Q = 0x040  # force only *memory* to be 64 bit.
OP_MEM_DQ = 0x050  # force only *memory* to be 128 bit.
OP_MEM_QQ = 0x060  # force only *memory* to be 256 bit.
OP_MEMMASK = 0x070  # this forces the memory to be a different size than the register. Reaches into OP_EXTRA_MEMSIZES

OP_NOVEXL = 0x080  # don't apply VEX.L here (even though it's set). TLDR: always 128/xmm reg

OP_EXTRA_MEMSIZES = [None, 1, 2, 4, 8, 16, 32]

OP_UNK = 0x000
OP_REG = 0x100
OP_IMM = 0x200
OP_REL = 0x300
OP_ADDR = 0x400
OP_EXPR = 0x500
OP_PTR = 0x600
OP_OFF = 0x700

OP_SIGNED = 0x001000
OP_STRING = 0x002000
OP_CONST = 0x004000
OP_NOREX = 0x008000

ARG_NONE = 0
cpu_8086 = 0x00001000
cpu_80286 = 0x00002000
cpu_80386 = 0x00003000
cpu_80387 = 0x00004000
cpu_80486 = 0x00005000
cpu_PENTIUM = 0x00006000
cpu_PENTPRO = 0x00007000
cpu_PENTMMX = 0x00008000
cpu_PENTIUM2 = 0x00009000
cpu_AMD64 = 0x0000a000
cpu_AESNI = 0x0000b000
cpu_AVX   = 0x0000c000
cpu_BMI   = 0x0000d000
cpu_OSPKE = 0x0000e000
cpu_VIRTUALPC = 0x0000f000

#eventually, change this for your own codes
#ADDEXP_SCALE_OFFSET= 0 
#ADDEXP_INDEX_OFFSET= 8
#ADDEXP_BASE_OFFSET = 16
#ADDEXP_DISP_OFFSET = 24
#MODRM_EA =  1
#MODRM_reg=  0
ADDRMETH_MASK = 0x00FF0000
OPTYPE_MASK = 0xFF000000
OPFLAGS_MASK = 0x0000FFFF

# NOTE: some notes from the intel manual...
# REX.W overrides 66, but alternate registers (via REX.B etc..) can have 66 to be 16 bit..
# REX.R only modifies reg for GPR/SSE(SIMD)/ctrl/debug addressing modes.
# REX.X only modifies the SIB index value
# REX.B modifies modrm r/m field, or SIB base (if SIB present), or opcode reg.
# We inherit all the regular intel prefixes...
# VEX replaces REX, and mixing them is invalid
PREFIX_REX   = 0x100000  # Shows that the rex prefix is present
PREFIX_REX_B = 0x010000  # Bit 0 in REX prefix (0x41) means ModR/M r/m field, SIB base, or opcode reg
PREFIX_REX_X = 0x020000  # Bit 1 in REX prefix (0x42) means SIB index extension
PREFIX_REX_R = 0x040000  # Bit 2 in REX prefix (0x44) means ModR/M reg extention
PREFIX_REX_W = 0x080000  # Bit 3 in REX prefix (0x48) means 64 bit operand
PREFIX_REX_MASK = PREFIX_REX_B | PREFIX_REX_X | PREFIX_REX_W | PREFIX_REX_R
PREFIX_REX_RXB  = PREFIX_REX_B | PREFIX_REX_X | PREFIX_REX_R



VMCS_NAMES = {
    0x0400: 'Guest_ES_Selector',
    0x0402: 'Guest_CS_Selector',
    0x0404: 'Guest_SS_Selector',
    0x0406: 'Guest_DS_Selector',
    0x0408: 'Guest_FS_Selector',
    0x040a: 'Guest_GS_Selector',
    0x040c: 'Guest_LDTR_Selector',
    0x040e: 'Guest_TR_Selector',

    0x0800: 'Host_ES_Selector',
    0x0802: 'Host_CS_Selector',
    0x0804: 'Host_SS_Selector',
    0x0806: 'Host_DS_Selector',
    0x0808: 'Host_FS_Selector',
    0x080a: 'Host_GS_Selector',
    0x080c: 'Host_TR_Selector',

    0x2000: 'Address_of_IO_bitmap_A_full',
    0x2001: 'Address_of_IO_bitmap_A_HIGH',
    0x2002: 'Address_of_IO_bitmap_B_full',
    0x2003: 'Address_of_IO_bitmap_B_HIGH',
    0x2004: 'Address_of_MSR_bitmaps_full',
    0x2005: 'Address_of_MSR_bitmaps_HIGH',
    0x2006: 'VM-exit_MSR_store_address_full',
    0x2007: 'VM-exit_MSR_store_address_HIGH',
    0x2008: 'VM-exit_MSR_load_address_full',
    0x2009: 'VM-exit_MSR_load_address_HIGH',
    0x200a: 'VM-entry_MSR_load_address_full',
    0x200b: 'VM-entry_MSR_load_address_HIGH',
    0x200c: 'Executive-VMCS_pointer_full',
    0x200d: 'Executive-VMCS_pointer_HIGH',

    0x2010: 'TSC_offset_full',
    0x2011: 'TSC_offset_HIGH',
    0x2012: 'Virtual_APIC_address_full',
    0x2013: 'Virtual_APIC_address_HIGH',
    0x2014: 'APIC-access_address_full',
    0x2015: 'APIC-access_address_HIGH',

    0x2800: 'VMCS_link_pointer_full',
    0x2801: 'VMCS_link_pointer_HIGH',
    0x2802: 'Guest_IA32_DEBUGCTL_full',
    0x2803: 'Guest_IA32_DEBUGCTL_HIGH',

    0x4000: 'Pin -based_VM-execution_controls',
    0x4002: 'Primary_process-based_VM-execution_controls',
    0x4004: 'Exception_bitmap',
    0x4006: 'Page_fault_error-code_mask',
    0x4008: 'Page-fault_error-code_match',
    0x400a: 'CR3-target_count',
    0x400c: 'VM-exit_controls',
    0x400e: 'VM-exit_MSR-store_count',
    0x4010: 'VM-exit_MSR-load_count',
    0x4012: 'VM-entry_controls',
    0x4014: 'VM-entry_MSR-load_count',
    0x4016: 'VM-entry_interruption-information_field',
    0x4018: 'VM-entry_exception_error_code',
    0x401a: 'VM-entry_instruction_length',
    0x401c: 'TPR_threshold',
    0x401e: 'Secondary_processor-based_VM-execution_controls',
    0x4020: 'PLE_Gap',
    0x4022: 'PLE_Window',

    0x4400: 'VM-instruction_error',
    0x4402: 'Exit_reason',
    0x4404: 'VM-exit_interruption_information',
    0x4406: 'VM-exit_interruption_error_code',
    0x4408: 'IDT-vectoring_information_field',
    0x440a: 'IDT-vectoring_error_code',
    0x440c: 'VM-exit_instruction_length',
    0x440e: 'VM-exit_instruction_information',

    0x4800: 'Guest_ES_limit',
    0x4802: 'Guest_CS_limit',
    0x4804: 'Guest_SS_limit',
    0x4806: 'Guest_DS_limit',
    0x4808: 'Guest_FS_limit',
    0x480a: 'Guest_GS_limit',
    0x480c: 'Guest_LDTR_limit',
    0x480e: 'Guest_TR_limit',
    0x4810: 'Guest_GDTR_limit',
    0x4812: 'Guest_IDTR_limit',
    0x4814: 'Guest_ES_access_rights',
    0x4816: 'Guest_CS_access_rights',
    0x4818: 'Guest_SS_access_rights',
    0x481a: 'Guest_DS_access_rights',
    0x481c: 'Guest_FS_access_rights',
    0x481e: 'Guest_GS_access_rights',
    0x4820: 'Guest_LDTR_access_rights',
    0x4822: 'Guest_TR_access_rights',
    0x4824: 'Guest_Interruptability_state',
    0x4826: 'Guest_activity_state',
    0x4828: 'Guest_SMBASE',
    0x482a: 'Guest_IA32_SYSENTER_CS',

    0x4c00: 'Host_IA32_SYSENTER_CS',

    0x6000: 'CR0_guest_host_mask',
    0x6002: 'CR4_guest_host_mask',
    0x6004: 'CR0_read_shadow',
    0x6006: 'CR4_read_shadow',
    0x6008: 'CR3_target_value_0',
    0x600a: 'CR3_target_value_1',
    0x600c: 'CR3_target_value_2',
    0x600e: 'CR3_target_value_3',

    0x6400: 'Exit_qualification',
    0x6402: 'IO_RCX',
    0x6404: 'IO_RSI',
    0x6406: 'IO_RDI',
    0x6408: 'IO_RIP',
    0x640a: 'Guest_linear_address',

    0x6800: 'Guest_CR0',
    0x6802: 'Guest_CR3',
    0x6804: 'Guest_CR4',
    0x6806: 'Guest_ES_base',
    0x6808: 'Guest_CS_base',
    0x680a: 'Guest_SS_base',
    0x680c: 'Guest_DS_base',
    0x680e: 'Guest_FS_base',
    0x6810: 'Guest_GS_base',
    0x6812: 'Guest_LDTR_base',
    0x6814: 'Guest_TR_base',
    0x6816: 'Guest_GDTR_base',
    0x6818: 'Guest_IDTR_base',
    0x681a: 'Guest_DR7',
    0x681c: 'Guest_RSP',
    0x681e: 'Guest_RIP',
    0x6820: 'Guest_RFLAGS',
    0x6822: 'Guest_pending_debug_exceptions',
    0x6824: 'Guest_IA32_SYSENTER_ESP',
    0x6826: 'Guest_IA32_SYSENTER_EIP',

    0x6800: 'Host_CR0',
    0x6802: 'Host_CR3',
    0x6804: 'Host_CR4',
    0x6806: 'Host_FS_base',
    0x6808: 'Host_GS_base',
    0x680a: 'Host_TR_base',
    0x680c: 'Host_GDTR_base',
    0x680e: 'Host_IDTR_base',
    0x6810: 'Host_IA32_SYSENTER_ESP',
    0x6812: 'Host_IA32_SYSENTER_EIP',
    0x6814: 'Host_RSP',
    0x6816: 'Host_RIP',
}



import struct
import logging
import traceback
import collections

import Elf
import Elf.elf_lookup as elf_lookup

import envi.exc as e_exc
import envi.bits as e_bits
import envi.const as e_const

import vivisect
import vivisect.parsers as v_parsers

from vivisect.const import *

from io import BytesIO


logger = logging.getLogger(__name__)

# FIXME: So this is terrible, but until we unite everything under one package
# namespace to rule them all, we need this to avoid a stupid amount of logspam
# but we can't just force things under
elog = logging.getLogger(Elf.__name__)
elog.setLevel(logger.getEffectiveLevel())


def parseFile(vw, filename, baseaddr=None):
    fd = open(filename, 'rb')
    elf = Elf.Elf(fd)
    return loadElfIntoWorkspace(vw, elf, filename=filename, baseaddr=baseaddr)

def parseBytes(vw, bytes, baseaddr=None):
    fd = BytesIO(bytes)
    elf = Elf.Elf(fd)
    return loadElfIntoWorkspace(vw, elf, baseaddr=baseaddr)

def parseFd(vw, fd, filename=None, baseaddr=None):
    fd.seek(0)
    elf = Elf.Elf(fd)
    return loadElfIntoWorkspace(vw, elf, filename=filename, baseaddr=baseaddr)

def parseMemory(vw, memobj, baseaddr):
    raise Exception('FIXME implement parseMemory for elf!')

def getMemBaseAndSize(vw, elf, baseaddr=None):
    '''
    Returns the default baseaddr and memory size required to load the file
    '''
    savebase = baseaddr

    memmaps = getMemoryMapInfo(elf)
    baseaddr = 0xffffffffffffffffffffffff
    topmem = 0

    for mapva, mperms, mname, mbytes, malign in memmaps:
        if mapva < baseaddr:
            baseaddr = mapva & -e_const.PAGE_SIZE   # align to page-size
        endva = mapva + len(mbytes)
        if endva > topmem:
            topmem = endva

    size = topmem - baseaddr

    if baseaddr == 0 and not elf.isRelocatable():
        baseaddr = vw.config.viv.parsers.elf.baseaddr
        
    if savebase:
        # if we provided a baseaddr, override what the file wants
        baseaddr = savebase
        
    return baseaddr, size



def getMemoryMapInfo(elf, fname=None, baseaddr=None):
    '''
    Gets the default baseaddr and memory map information
    All the information necessary to add memory maps (or get overall size info)
    '''
    memmaps = []

    addbase, baseoff, baseaddr = getAddBaseAddr(elf, baseaddr)

    pgms = elf.getPheaders()
    for pgm in pgms:
        if pgm.p_type == Elf.PT_LOAD:
            if pgm.p_memsz == 0:
                continue
            logger.info('Loading: %s', pgm)
            bytez = elf.readAtOffset(pgm.p_offset, pgm.p_filesz)
            bytez += b'\x00' * (pgm.p_memsz - pgm.p_filesz)
            pva = pgm.p_vaddr
            pva += baseoff
            memmaps.append((pva, pgm.p_flags & 0x7, fname, bytez, e_const.PAGE_SIZE))
        else:
            logger.info('Skipping: %s', pgm)

    if len(pgms) == 0:
        # NOTE: among other file types, Linux Kernel modules have no program headers
        secs = elf.getSections()
        # fall back to loading sections as best we can...
        logger.info('elf: no program headers found! (in %r)', fname)

        maps = [ [s.sh_offset,s.sh_size,s.sh_addralign] for s in secs if s.sh_offset and s.sh_size ]
        maps.sort()

        merged = []
        for i in range(len(maps)):

            if merged and maps[i][0] == (merged[-1][0] + merged[-1][1]):
                merged[-1][1] += maps[i][1]
                continue

            merged.append( maps[i] )

        for offset,size,align in merged:
            bytez = elf.readAtOffset(offset,size)
            memmaps.append((baseaddr + offset, 0x7, fname, bytez, align))

        for sec in secs:
            if sec.sh_offset and sec.sh_size:
                sec.sh_addr = baseaddr + sec.sh_offset

    return memmaps



def makeStringTable(vw, va, maxva):

    while va < maxva:
        if vw.readMemory(va, 1) == "\x00":
            va += 1
            continue
        else:
            try:
                if vw.isLocation(va):
                    return
                l = vw.makeString(va)
                va += l[vivisect.L_SIZE]
            except Exception as e:
                logger.warning("makeStringTable\t%r", e)
                return

def makeSymbolTable(vw, va, maxva):
    ret = []
    sname = 'elf.Elf%dSymbol' % (vw.getPointerSize() * 8)
    while va < maxva:
        s = vw.makeStructure(va, sname)
        ret.append(s)
        va += len(s)
    return ret

def makeDynamicTable(vw, va, maxva, baseoff=0):
    logger.debug("0x%x -> 0x%x", va, maxva)
    ret = []
    sname = 'elf.Elf%dDynamic' % (vw.getPointerSize() * 8)
    while va < maxva:
        s = vw.makeStructure(va, sname)
        ret.append(s)
        dtag = s.d_tag
        dtype = Elf.dt_types.get(dtag, "Unknown Dynamic Type")
        vw.setComment(va, dtype)
        va += len(s)
        if dtag == Elf.DT_NULL:
            break
    return ret

def makeRelocTable(vw, va, maxva, baseoff, addend=False):
    if addend:
        sname = 'elf.Elf%dReloca' % (vw.getPointerSize() * 8)
    else:
        sname = 'elf.Elf%dReloc' % (vw.getPointerSize() * 8)

    while va < maxva:
        s = vw.makeStructure(va, sname)
        tname = Elf.r_types_386.get(Elf.getRelocType(s.r_info), "Unknown Type")
        vw.setComment(va, tname)
        va += len(s)

def makeFunctionTable(elf, vw, tbladdr, size, tblname, funcs, ptrs, baseoff=0):
    logger.debug('makeFunctionTable(tbladdr=0x%x, size=0x%x, tblname=%r,  baseoff=0x%x)', tbladdr, size, tblname, baseoff)
    psize = vw.getPointerSize()

    # If the provided buffer is shorter than the standard pointer size, use the
    # buffer length
    if size < psize:
        psize = size

    fmtgrps = e_bits.fmt_chars[vw.getEndian()]
    pfmt = fmtgrps[psize]

    if size % psize != 0:
        # If the table size isn't a multiple of the pointer size, round down.
        # It doesn't really make sense for this to be the case, 
        # but this is seen in a small number of real world samples.
        size -= (size % psize)

    secbytes = elf.readAtRva(tbladdr, size)
    tbladdr += baseoff

    ptr_count = 0
    for off in range(0, size, psize):
        addr, = struct.unpack_from(pfmt, secbytes, off)
        addr += baseoff

        nstub = tblname + "_%d"
        pname = nstub % ptr_count

        vw.makeName(addr, pname, filelocal=True, makeuniq=True)
        ptrs.append((tbladdr + off, addr, pname))
        funcs.append((pname, addr))
        ptr_count += 1


arch_names = {
    Elf.EM_ARM: 'arm',
    Elf.EM_386: 'i386',
    Elf.EM_X86_64: 'amd64',
    Elf.EM_MSP430: 'msp430',
    Elf.EM_ARM_A64: 'a64',
}

archcalls = {
    'i386': 'cdecl',
    'amd64': 'sysvamd64call',
    'arm': 'armcall',
    'thumb': 'armcall',
    'thumb16': 'armcall',
    'a64': 'a64call',
}

def getAddBaseAddr(elf, baseaddr=None):
    '''
    # NOTE: This is only for prelink'd so's and exe's.  Make something for old style so.
    '''
    baseoff = 0
    logger.debug("baseaddr: %r\tbaseoff: 0x%x", baseaddr, baseoff)
    elfbaseaddr = elf.getBaseAddress()
    if baseaddr is None:
        baseaddr = elfbaseaddr

    if not elf.isPreLinked() and elf.isSharedObject():
        addbase = True
        baseoff = baseaddr

    elif elf.isRelocatable():
        addbase = True
        baseoff = 0

    else:
        addbase = False
        baseoff = baseaddr - elfbaseaddr

    logger.debug("baseaddr: %r\tbaseoff: 0x%x", baseaddr, baseoff)
    return addbase, baseoff, baseaddr

def loadElfIntoWorkspace(vw, elf, filename=None, baseaddr=None):
    # analysis of discovered functions and data locations should be stored until the end of loading
    logger.info("loadElfIntoWorkspace(filename=%r, baseaddr: %r", filename, baseaddr)
    data_ptrs = []
    new_pointers = []
    new_functions = []

    # get baseaddr and size, then make sure we have a good baseaddr
    filebaseaddr, size = getMemBaseAndSize(vw, elf, baseaddr=baseaddr)
    if baseaddr is None:
        baseaddr = filebaseaddr

    logger.debug('initial file baseva: 0x%x  size: 0x%x', baseaddr, size)
    baseaddr = vw.findFreeMemoryBlock(size, baseaddr)
    logger.debug("loading %r (size: 0x%x) at 0x%x", filename, size, baseaddr)

    arch = arch_names.get(elf.e_machine)
    if isinstance(arch, dict):
        arch = arch.get(elf.bits)
    if arch is None:
        raise Exception("Unsupported Architecture: %d (%d bits)\n" % (elf.e_machine, elf.bits))

    platform = elf.getPlatform()

    # setup needed platform/format
    vw.setMeta('Architecture', arch)
    vw.setMeta('Platform', platform)
    vw.setMeta('Format', 'elf')
    vw.parsedbin = elf
    # Treat the event system veeery carefully
    byts = elf.getFileBytes()
    vw.setMeta('FileBytes', v_parsers.compressBytes(byts))

    vw.setMeta('DefaultCall', archcalls.get(arch,'unknown'))

    vw.addNoReturnApi("*.abort")
    vw.addNoReturnApi("*.exit")
    vw.addNoReturnApi("*._exit")
    vw.addNoReturnApi("*.longjmp")
    vw.addNoReturnApi("*._setjmp")
    vw.addNoReturnApi("*.j__ZSt9terminatev")
    vw.addNoReturnApi("*.std::terminate(void)")
    vw.addNoReturnApi("*.__assert_fail")
    vw.addNoReturnApi("*.__stack_chk_fail")
    vw.addNoReturnApi("*.pthread_exit")
    vw.addNoReturnApi("*.longjmp")

    # for VivWorkspace, MSB==1, LSB==0... which is the same as True/False for Big-Endian
    vw.setEndian(elf.getEndian())

    # Base addr is earliest section address rounded to pagesize
    # Some ELF's require adding the baseaddr to most/all later addresses
    addbase, baseoff, baseaddr = getAddBaseAddr(elf, baseaddr)
    logger.debug("Loading ELF into workspace: addbase: %r, baseoff: 0x%x, baseaddr: 0x%x", addbase, baseoff, baseaddr)

    # Keep track of if a LOAD happens for file offset 0
    elfHdrAtOffset0 = False

    elf.fd.seek(0)
    md5hash = v_parsers.md5Bytes(byts)
    sha256 = v_parsers.sha256Bytes(byts)

    if filename is None:
        # see if dynamics DT_SONAME holds a name for us
        for dyn in elf.getDynamics():
            if dyn.d_tag == Elf.DT_SONAME:
                filename = dyn.getName()

        # if all else fails, fallback
        if filename is None:
            filename = "elf_%.8x" % baseaddr

    fname = vw.addFile(filename.lower(), baseaddr, md5hash)
    vw.setFileMeta(fname, 'sha256', sha256)
    vw.setFileMeta(fname, 'relro', getRelRo(elf))
    vw.setFileMeta(fname, 'canaries', hasStackCanaries(vw))
    vw.setFileMeta(fname, 'nx', hasNX(elf))
    vw.setFileMeta(fname, 'pie', hasPIE(elf))
    vw.setFileMeta(fname, 'rpath', hasRPATH(elf))
    vw.setFileMeta(fname, 'runpath', hasRUNPATH(elf))
    vw.setFileMeta(fname, 'stripped', isStripped(elf))

    secnames = []
    for sec in elf.getSections():
        secnames.append(sec.getName())

    secs = elf.getSections()

    for mmapva, mmperms, mfname, mbytez, malign in getMemoryMapInfo(elf, fname, baseaddr):
        logger.debug("vw.addMemoryMap(0x%x, 0x%x, %r, 0x%x, 0x%x)", mmapva, mmperms, mfname, len(mbytez), malign)
        vw.addMemoryMap(mmapva, mmperms, mfname, mbytez, malign)

    # First add all section definitions so we have them
    for sec in secs:
        sname = sec.getName()
        size = sec.sh_size
        if sec.sh_addr == 0:
            continue # Skip non-memory mapped sections

        sva = sec.sh_addr
        sva += baseoff

        vw.addSegment(sva, size, sname, fname)

    # since getFileByVa is based on segments, and ELF Sections seldom cover all the
    # loadable memory space.... we'll add PT_LOAD Program Headers, only at the
    # end.  If we add them first, they're always the matching segments.  At the 
    # end, they make more of a default segment
    pcount = 0
    if vw.getFileByVa(baseaddr) is None:
        for phdr in elf.getPheaders():
            if phdr.p_type != Elf.PT_LOAD:
                continue

            sva = phdr.p_vaddr
            sva += baseoff

            vw.addSegment(sva, phdr.p_memsz, 'PHDR%d' % pcount, fname)
            pcount += 1

    # load information from dynamics:
    f_init = elf.dyns.get(Elf.DT_INIT)
    if f_init is not None:
        f_init += baseoff
        vw.makeName(f_init, "init_function", filelocal=True, makeuniq=True)
        vw.addEntryPoint(f_init)

    f_fini = elf.dyns.get(Elf.DT_FINI)
    if f_fini is not None:
        f_fini += baseoff
        vw.makeName(f_fini, "fini_function", filelocal=True, makeuniq=True)
        vw.addEntryPoint(f_fini)

    f_inita = elf.dyns.get(Elf.DT_INIT_ARRAY)
    if f_inita is not None:
        f_initasz = elf.dyns.get(Elf.DT_INIT_ARRAYSZ)
        makeFunctionTable(elf, vw, f_inita, f_initasz, 'init_array', new_functions, new_pointers, baseoff)

    f_finia = elf.dyns.get(Elf.DT_FINI_ARRAY)
    if f_finia is not None:
        f_finiasz = elf.dyns.get(Elf.DT_FINI_ARRAYSZ)
        makeFunctionTable(elf, vw, f_finia, f_finiasz, 'fini_array', new_functions, new_pointers, baseoff)

    f_preinita = elf.dyns.get(Elf.DT_PREINIT_ARRAY)
    if f_preinita is not None:
        f_preinitasz = elf.dyns.get(Elf.DT_PREINIT_ARRAYSZ)
        makeFunctionTable(elf, vw, f_preinita, f_preinitasz, 'preinit_array', new_functions, new_pointers, baseoff)

    # dynamic table
    phdr = elf.getDynPHdr()     # this is the Program Header which points at the DYNAMICS table, not the other way around.
    if phdr is not None:
        sva, size = phdr.p_vaddr, phdr.p_memsz
        sva += baseoff     # getDynInfo returns (offset, filesz)
        makeDynamicTable(vw, sva, sva+size, baseoff)
        # if there's no Dynamics PHDR, don't bother trying to parse it from Sections.  It doesn't exist.

    # dynstr table
    sva, size = elf.getDynStrTabInfo()
    if sva is not None:
        sva += baseoff
        makeStringTable(vw, sva, sva+size)

    # dynsyms table
    sva, symsz, size = elf.getDynSymTabInfo()
    if sva is not None:
        sva += baseoff
        logger.debug("making DynSym table: 0x%x secsize: 0x%x entsize: 0x%x", sva, size, symsz)
        [s for s in makeSymbolTable(vw, sva, sva+size)]

    # Now trigger section specific analysis
    # Test to make sure Dynamics info is king.  Sections info should not overwrite Dynamics
    for sec in secs:
        sname = sec.getName()
        size = sec.sh_size
        if sec.sh_addr == 0:
            continue # Skip non-memory mapped sections

        sva = sec.sh_addr
        sva += baseoff

        # if we've already defined a location at this address, skip it. (eg. DYNAMICS)
        if vw.getLocation(sva) == sva:
            continue

        if sname == ".interp":
            vw.makeString(sva)

        elif sname == ".init":
            init_tup = ("init_function", sva)
            if init_tup not in new_functions:
                vw.makeName(sva, "init_function", filelocal=True, makeuniq=True)
                new_functions.append(init_tup)

        elif sname == ".init_array":
            makeFunctionTable(elf, vw, sec.sh_addr, size, 'init_function', new_functions, new_pointers, baseoff)

        elif sname == ".fini":
            fini_tup = ("fini_function", sva)
            if fini_tup not in new_functions:
                vw.makeName(sva, "fini_function", filelocal=True, makeuniq=True)
                new_functions.append(fini_tup)

        elif sname == ".fini_array":
            makeFunctionTable(elf, vw, sec.sh_addr, size, 'fini_function', new_functions, new_pointers, baseoff)

        elif sname == ".dynamic":  # Imports
            makeDynamicTable(vw, sva, sva+size, baseoff)

        elif sname == ".dynstr":  # String table for dynamics
            makeStringTable(vw, sva, sva+size)

        elif sname == ".dynsym":
            [s for s in makeSymbolTable(vw, sva, sva+size)]

        # If the section is really a string table, do it
        if sec.sh_type == Elf.SHT_STRTAB:
            makeStringTable(vw, sva, sva+size)

        elif sec.sh_type == Elf.SHT_SYMTAB:
            makeSymbolTable(vw, sva, sva+size)

        elif sec.sh_type == Elf.SHT_REL:
            makeRelocTable(vw, sva, sva+size, baseoff)

        elif sec.sh_type == Elf.SHT_RELA:
            makeRelocTable(vw, sva, sva+size, baseoff, addend=True)

        if sec.sh_flags & Elf.SHF_STRINGS:
            makeStringTable(vw, sva, sva+size)

    # get "Dynamics" based items, like NEEDED libraries (dependencies)
    elfmeta = {}
    for d in elf.getDynamics():
        if d.d_tag == Elf.DT_NEEDED:
            name = d.getName()
            name = name.split('.')[0].lower()
            vw.addLibraryDependancy(name)
        else:
            logger.debug("DYNAMIC:\t%r", d)
        
        dval = d.d_value
        if d.d_tag in Elf.dt_rebase and addbase:
            dval += baseoff

        elfmeta[Elf.dt_names.get(d.d_tag)] = dval

    # TODO: create a VaSet instead? setMeta allows more free-form info,
    # but isn't currently accessible from the gui
    vw.setFileMeta(fname, 'ELF_DYNAMICS', elfmeta)
    vw.setFileMeta(fname, 'addbase', addbase)
    vw.setFileMeta(fname, 'baseoff', baseoff)

    # applyRelocs is specifically prior to "process Dynamic Symbols" because Dynamics-only symbols
    # (ie. not using Section Headers) may not get all the symbols.  Some ELF's simply list too
    # small a space using SYMTAB and SYMTABSZ
    postfix = applyRelocs(elf, vw, addbase, baseoff)

    # process Dynamic Symbols - this must happen *after* relocations, which can expand the size of this
    for s in elf.getDynSyms():
        stype = s.getInfoType()
        sva = s.st_value

        if sva == 0:
            continue

        sva += baseoff
        if sva == 0:
            continue

        dmglname = demangle(s.name)

        if stype == Elf.STT_FUNC or \
                (stype == Elf.STT_GNU_IFUNC and arch in ('i386', 'amd64')):   # HACK: linux is what we're really after.
            try:
                new_functions.append(("DynSym: STT_FUNC", sva))
                vw.addExport(sva, EXP_FUNCTION, dmglname, fname, makeuniq=True)
                vw.setComment(sva, s.name)
            except Exception as e:
                vw.vprint('addExport Failure: (%s) %s' % (s.name, e))

        elif stype == Elf.STT_OBJECT:
            if vw.isValidPointer(sva):
                try:
                    vw.addExport(sva, EXP_DATA, dmglname, fname, makeuniq=True)
                    vw.setComment(sva, s.name)
                except Exception:
                    vw.vprint('STT_OBJECT Warning: %s' % traceback.format_exc())

        elif stype == Elf.STT_HIOS:
            # So aparently Elf64 binaries on amd64 use HIOS and then
            # s.st_other cause that's what all the kewl kids are doing...
            sva = s.st_other
            sva += baseoff
            if vw.isValidPointer(sva):
                try:
                    new_functions.append(("DynSym: STT_HIOS", sva))
                    vw.addExport(sva, EXP_FUNCTION, dmglname, fname, makeuniq=True)
                    vw.setComment(sva, s.name)
                except Exception:
                    vw.vprint('STT_HIOS Warning:\n%s' % traceback.format_exc())

        elif stype == Elf.STT_MDPROC:    # there's only one that isn't HI or LO...
            sva = s.st_other
            sva += baseoff
            if vw.isValidPointer(sva):
                try:
                    vw.addExport(sva, EXP_DATA, dmglname, fname, makeuniq=True)
                    vw.setComment(sva, s.name)
                except Exception:
                    vw.vprint('STT_MDPROC Warning:\n%s' % traceback.format_exc())

        else:
            logger.debug("DYNSYM:\t%r\t%r\t%r\t%r", s, s.getInfoType(), 'other', hex(s.st_other))

        if dmglname in postfix:
            for rlva, addend in postfix[dmglname]:
                vw.addRelocation(rlva, RTYPE_BASEPTR, sva + addend - baseoff)

    vaSetNames = vw.getVaSetNames()
    if not 'FileSymbols' in vaSetNames:
        vw.addVaSet("FileSymbols", (("Name", VASET_STRING), ("va", VASET_ADDRESS)))
    if not 'WeakSymbols' in vaSetNames:
        vw.addVaSet("WeakSymbols", (("Name", VASET_STRING), ("va", VASET_ADDRESS)))

    # apply symbols to workspace (if any)
    relocs = [r for idx, r in elf.getRelocs()]
    impvas = [va for va, x, y, z in vw.getImports()]
    expvas = [va for va, x, y, z in vw.getExports()]
    for s in elf.getSymbols():
        sva = s.st_value
        shndx = s.st_shndx
        if elf.isRelocatable():
            if shndx == elf_lookup.SHN_ABS:
                pass
            elif shndx in elf_lookup.shn_special_section_indices:
                # TODO: We should do things with SHN_ABS
                pass
            else:
                section = elf.getSectionByIndex(shndx)
                if section:
                    sva += section.sh_addr

        dmglname = demangle(s.name)
        logger.debug('symbol val: 0x%x\ttype: %r\tbind: %r\t name: %r', sva,
                                                                        Elf.st_info_type.get(s.getInfoType(), s.st_info),
                                                                        Elf.st_info_bind.get(s.getInfoBind(), s.st_other),
                                                                        s.name)

        symtype = s.getInfoType()
        if symtype == Elf.STT_FILE:
            vw.setVaSetRow('FileSymbols', (dmglname, sva))
            continue

        elif symtype == Elf.STT_NOTYPE:
            # mapping symbol
            if arch in ('arm', 'thumb', 'thumb16'):
                symname = s.getName()
                sva += baseoff
                if symname == '$a':
                    # ARM code
                    logger.info('mapping (NOTYPE) ARM symbol: 0x%x: %r', sva, dmglname)
                    new_functions.append(("Mapping Symbol: $a", sva))

                elif symname == '$t':
                    # Thumb code
                    logger.info('mapping (NOTYPE) Thumb symbol: 0x%x: %r', sva, dmglname)
                    new_functions.append(("Mapping Symbol: $t", sva + 1))

                elif symname == '$d':
                    # Data Items (eg. literal pool)
                    logger.info('mapping (NOTYPE) data symbol: 0x%x: %r', sva, dmglname)
                    data_ptrs.append(sva)
        elif symtype == Elf.STT_OBJECT:
            # "This symbol is associated with a data object, such as a variable, an array, and so forth."

            symname = s.getName()
            sva += baseoff
            if symname:
                vw.makeName(sva, symname, filelocal=True, makeuniq=True)

            if sva == 0x0:
                # if the address of the symbol is 0x0,
                # then it is probably an extern object, like the environ pointer provided by glibc.
                logger.debug("STT_OBJECT symbol %s@0x%x has address 0x0: skipping.", 
                             symname or "",
                             sva)
            elif not vw.isValidPointer(sva):
                # if the address of the symbol is invalid, 
                # this is weird, but we'll try to carry on.
                logger.debug("STT_OBJECT symbol %s@0x%x has invalid address: skipping.", 
                             symname or "",
                             sva)
            else:
                # the address of the symbol is valid.
                # the data named by the symbol is in this image.
                # so we'll try to interpret the data there in the following ways:
                #
                #   1. as an empty object
                #   2. as Unicode string
                #   3. as ASCII string
                #   4. as a pointer
                #   5. as an array of pointers
                #   6. as a number

                if s.st_size == 0:
                    # via Oracle docs:
                    # > Many symbols have associated sizes.
                    # > For example, a data object's size is the number of bytes that are contained in the object.
                    # > This member holds the value zero if the symbol has no size or an unknown size.
                    #
                    # Let's assume there's at least a pointer-sized object here.
                    valu = vw.readMemoryPtr(sva)
                    if valu == 0x0:
                        # this is either the number 0,
                        # or a null-initialized pointer, like `void *foo = NULL;`
                        # without analyzing how this data is used, we can't tell.
                        # so assume its a number for now.
                        logger.debug("STT_OBJECT symbol %s@0x%x with zero size and NULL value: guessing number 0x0", 
                                     symname or "",
                                     sva)
                        vw.makeNumber(sva, size=vw.getPointerSize())
                    elif vw.isValidPointer(valu):
                        logger.debug("STT_OBJECT symbol %s@0x%x with zero size: guessing valid pointer",
                                     symname or "",
                                     sva)
                        new_pointers.append((sva, valu, symname or ""))
                    else:
                        logger.debug("STT_OBJECT symbol %s@0x%x with zero size: guessing number",
                                     symname or "",
                                     sva)
                        vw.makeNumber(sva, size=vw.getPointerSize())
                elif vw.isProbablyUnicode(sva):
                    vw.makeUnicode(sva, size=s.st_size)
                elif vw.isProbablyString(sva):
                    vw.makeString(sva, size=s.st_size)
                elif s.st_size == vw.getPointerSize():
                    # appears to be a single pointer, or number of pointer-width
                    valu = vw.readMemoryPtr(sva)
                    if valu == 0x0:
                        # this is either the number 0,
                        # or a null-initialized pointer, like `void *foo = NULL;`
                        # without analyzing how this data is used, we can't tell.
                        # so assume its a number for now.
                        logger.debug("STT_OBJECT symbol %s@0x%x with pointer size and NULL value: guessing number 0x0", 
                                     symname or "",
                                     sva)
                        vw.makeNumber(sva, size=vw.getPointerSize())
                    elif vw.isValidPointer(valu):
                        logger.debug("STT_OBJECT symbol %s@0x%x with pointer size: guessing valid pointer", 
                                     symname or "",
                                     sva)
                        new_pointers.append((sva, valu, symname or ""))
                    else:
                        logger.debug("STT_OBJECT symbol %s@0x%x with pointer size: guessing number", 
                                     symname or "",
                                     sva)
                        vw.makeNumber(sva, size=vw.getPointerSize())
                elif s.st_size % vw.getPointerSize() == 0 and s.st_size > vw.getPointerSize():
                    # because this object size is pointer-aligned,
                    # this might be an array of pointers, or a structure with embedded pointers.
                    # so walk through the entries (pointer-aligned) and try to extract pointers.
                    for addr in range(sva, sva+s.st_size, vw.psize):
                        valu = vw.readMemoryPtr(addr)
                        if vw.isValidPointer(valu):
                            new_pointers.append((addr, valu, symname or ""))
                elif s.st_size in (1, 2, 4, 8, 16):
                    logger.debug("STT_OBJECT symbol %s@0x%x: guessing number", 
                                 symname or "",
                                 sva)
                    # we don't really know how to interpret the data, 
                    # so fall back to a number.
                    vw.makeNumber(sva, size=s.st_size)
                else:
                    # we really don't know what to do, and thats ok.
                    logger.warning("STT_OBJECT symbol %s@0x%x with unaligned size: unknown data", 
                                   symname or "",
                                   sva)

        # if the symbol has a value of 0, it is likely a relocation point which gets updated
        sname = demangle(s.name)
        if sva == 0:
            for reloc in relocs:
                rname = demangle(reloc.name)
                if rname == sname:
                    sva = reloc.r_offset
                    logger.info('sva==0, using relocation name: %x: %r', sva, rname)
                    break

        dmglname = demangle(sname)

        # TODO: make use of _ZTSN (typeinfo) and _ZTVN (vtable) entries if name demangles cleanly

        sva += baseoff
        if vw.isValidPointer(sva) and len(dmglname):
            try:
                if s.getInfoBind() == Elf.STB_WEAK:
                    logger.info('WEAK symbol: 0x%x: %r', sva, sname)
                    vw.setVaSetRow('WeakSymbols', (sname, sva))
                    dmglname = '__weak_' + dmglname

                if sva in impvas or sva in expvas:
                    imps = [imp for imp in vw.getImports() if imp[0] == sva]
                    exps = [exp for exp in vw.getExports() if exp[0] == sva]
                    logger.debug('skipping Symbol naming for existing Import/Export: 0x%x (%r) (%r) (%r)', sva, s.name, imps, exps)
                else:
                    vw.makeName(sva, dmglname, filelocal=True, makeuniq=True)

            except Exception as e:
                logger.warning("%s" % str(e))

        if s.getInfoType() == Elf.STT_FUNC:
            new_functions.append(("STT_FUNC", sva))

    eentry = baseoff + elf.e_entry

    if vw.isValidPointer(eentry):
        vw.addExport(eentry, EXP_FUNCTION, '__entry', fname)
        new_functions.append(("ELF Entry", eentry))

    if elfHdrAtOffset0 and vw.isValidPointer(baseaddr):
        sname = 'elf.Elf%d' % (vw.getPointerSize() * 8)
        vw.makeStructure(baseaddr, sname)



    # mark all the entry points for analysis later
    for cmnt, fva in new_functions:
        # If the address of the potential new function is the ELF base address
        # (therefore something that had an offset of 0, and points to the ELF
        # header itself), skip it
        if fva == baseaddr:
            # throw the null offset functions into a VAset
            vw.setVaSetRow("Null Offset Functions", (cmnt, fva))
            logger.warning("Skipping 'new' function pointing at baseaddr: 0x%x (%s)", fva, cmnt)
            continue

        logger.info('adding function from ELF metadata: 0x%x (%s)', fva, cmnt)
        vw.addEntryPoint(fva)   # addEntryPoint queue's code analysis for later in the analysis pass

    # mark all the pointers for analysis later
    for va, tva, pname in new_pointers:
        logger.info('adding pointer 0x%x -> 0x%x', va, tva)
        vw.setVaSetRow('PointersFromFile', (va, tva, fname, pname))

    return fname


def applyRelocs(elf, vw, addbase=False, baseoff=0):
    '''
    process relocations / strings (relocs use Dynamic Symbols)
    '''
    postfix = collections.defaultdict(list)
    arch = arch_names.get(elf.e_machine)
    othr = None
    for secidx, r in elf.getRelocs():
        rtype = r.getType()
        rlva = r.r_offset
        if elf.isRelocatable():
            container = elf.getSectionByIndex(secidx)
            if container.sh_flags & elf_lookup.SHF_INFO_LINK:
                othr = elf.getSectionByIndex(container.sh_info)
                if othr:
                    rlva += othr.sh_addr

        rlva += baseoff

        try:
            # If it has a name, it's an externally resolved "import" entry,
            # otherwise, just a regular reloc
            name = r.getName()
            dmglname = demangle(name)
            logger.debug('relocs: 0x%x: %s (%s)', rlva, dmglname, name)
            if arch in ('i386', 'amd64'):
                if name:
                    #if dmglname == 
                    if rtype == Elf.R_X86_64_IRELATIVE:
                        # before making import, let's fix up the pointer as a BASEPTR Relocation
                        ptr = r.r_addend
                        rloc = vw.addRelocation(rlva, RTYPE_BASEPTR, ptr)
                        if rloc:
                            logger.info('Reloc: R_X86_64_IRELATIVE 0x%x', rlva)

                    if rtype in (Elf.R_386_JMP_SLOT, Elf.R_X86_64_GLOB_DAT, Elf.R_X86_64_IRELATIVE):
                        logger.info('Reloc: making Import 0x%x (name: %s/%s) ', rlva, name, dmglname)
                        vw.makeImport(rlva, "*", dmglname)
                        vw.setComment(rlva, name)

                    elif rtype == Elf.R_386_COPY:  # Also covers X86_64_COPY
                        # the linker is responsible for filling these in so we probably won't have these
                        vw.addRelocation(rlva, RTYPE_BASERELOC, 0)

                    elif rtype == Elf.R_386_32:  # Also covers X86_64_64
                        # a direct punch in plus an addend
                        # but things like libstc++ use this type for vtables in the rel.dyn
                        # section without actually specifying an addend
                        postfix[dmglname].append((rlva, getattr(r, 'r_addend', 0)))

                    else:
                        logger.warning('unknown reloc type: %d %s (at %s)', rtype, name, hex(rlva))
                        logger.info(r.tree())
                        vw.makeName(rlva, dmglname, makeuniq=True)
                        if name != dmglname:
                            vw.setComment(rlva, name)

                else:
                    if rtype == Elf.R_386_RELATIVE: # R_X86_64_RELATIVE is the same number
                        ptr = vw.readMemoryPtr(rlva)
                        logger.info('R_386_RELATIVE: adding Relocation 0x%x -> 0x%x (name: %s) ', rlva, ptr, dmglname)
                        vw.addRelocation(rlva, RTYPE_BASEPTR, ptr)

                    elif arch == 'amd64' and rtype == Elf.R_X86_64_32S:
                        # the same as R_386_32 except we don't always get a name.
                        ptr = r.r_addend
                        symbol = elf.getSymbols()[r.getSymTabIndex()]
                        valu = symbol.st_value
                        if symbol.getInfoType() == elf_lookup.STT_SECTION:
                            ref = elf.getSectionByIndex(symbol.st_shndx)
                            if ref:
                                valu = ref.sh_addr
                        vw.addRelocation(rlva, RTYPE_BASEOFF, data=symbol.st_value + ptr, size=4)

                    elif rtype == Elf.R_X86_64_IRELATIVE:
                        # first make it a relocation that is based on the imagebase
                        ptr = r.r_addend
                        logger.info('R_X86_64_IRELATIVE: adding Relocation 0x%x -> 0x%x (name: %r %r) ', rlva, ptr, name, dmglname)
                        rloc = vw.addRelocation(rlva, RTYPE_BASEPTR, ptr)
                        if rloc is not None:
                            continue

                        # next get the target and find a name, since the reloc itself doesn't have one
                        tgt = vw.readMemoryPtr(rlva)
                        tgtname = vw.getName(tgt)
                        if tgtname is not None:
                            logger.info('   name(0x%x): %r', tgt, tgtname)
                            fn, symname = tgtname.split('.', 1)
                            logger.info('Reloc: making Import 0x%x (name: %s.%s) ', rlva, fn, symname)
                            vw.makeImport(rlva, fn, symname)
                    # TODO: This isn't strictly wrong, but does induce graph building errors
                    #elif rtype == Elf.R_X86_64_PLT32:
                    #    # plt + addend- secoff
                    #    addend = r.r_addend
                    #    plt = elf.getSection('.plt')
                    #    if plt:
                    #        addend += plt.sh_addr
                    #    if othr:
                    #        addend -= othr.sh_addr
                    #    vw.addRelocation(rlva, RTYPE_BASEPTR, data=addend, size=4)
                    elif rtype == Elf.R_X86_64_TPOFF64:
                        pass
                    elif rtype == Elf.R_386_TLS_DTPMOD32:
                        pass
                    else:
                        logger.warning('unknown reloc type: %d %s (at %s)', rtype, name, hex(rlva))
                        logger.warning(r.tree())


            if arch in ('arm', 'thumb', 'thumb16',):
                # ARM REL entries require an addend that could be stored as a
                # number or an instruction!
                import envi.archs.arm.const as eaac
                if r.vsHasField('r_addend'):
                    # this is a RELA object, bringing its own addend field!
                    addend = r.r_addend
                else:
                    # otherwise, we have to check the stored value for number or instruction
                    # if it's an instruction, we have to use the immediate value and then
                    # figure out if it's negative based on the instruction!
                    try:
                        temp = vw.readMemoryPtr(rlva)
                        if rtype in Elf.r_armclasses[Elf.R_ARMCLASS_DATA] or rtype in Elf.r_armclasses[Elf.R_ARMCLASS_MISC]:
                            # relocation points to a DATA or MISCELLANEOUS location
                            addend = temp
                        else:
                            # relocation points to a CODE location (ARM, THUMB16, THUMB32)
                            op = vw.parseOpcode(rlva)
                            for oper in op.opers:
                                if hasattr(oper, 'val'):
                                    addend = oper.val
                                    break

                                elif hasattr(oper, 'offset'):
                                    addend = oper.offset
                                    break

                            lastoper = op.opers[-1]
                            if op.mnem.startswith('sub') or \
                                    op.mnem in ('ldr', 'str') and \
                                    hasattr(lastoper, 'pubwl') and \
                                    not (lastoper.pubwl & eaac.PUxWL_ADD):
                                        addend = -addend
                    except Exception as e:
                        logger.exception("ELF: Reloc Addend determination: %s", e)
                        addend = temp

                logger.debug('addend: 0x%x', addend)

                if rtype == Elf.R_ARM_JUMP_SLOT:
                    symidx = r.getSymTabIndex()
                    sym = elf.getDynSymbol(symidx)
                    ptr = sym.st_value

                    # quick check to make sure we don't provide this symbol
                    # some toolchains like to point the GOT back at it's PLT entry
                    # that does *not* mean it's not an IMPORT
                    if ptr and not isPLT(vw, ptr):
                        logger.info('R_ARM_JUMP_SLOT: adding Relocation 0x%x -> 0x%x (%s) ', rlva, ptr, dmglname)
                        # even if addRelocation fails, still make the name, same thing down in GLOB_DAT
                        if addbase:
                            vw.addRelocation(rlva, vivisect.RTYPE_BASEPTR, ptr)
                        else:
                            vw.addRelocation(rlva, vivisect.RTYPE_BASERELOC, ptr)
                        pname = "ptr_%s_%.8x" % (name, rlva)
                        if vw.vaByName(pname) is None:
                            vw.makeName(rlva, pname)

                        # name the target as well
                        ptr += baseoff
                        # normalize thumb addresses
                        ptr &= -2
                        vw.makeName(ptr, name)
                        vw.setComment(ptr, dmglname)

                    else:
                        logger.info('R_ARM_JUMP_SLOT: adding Import 0x%x (%s) ', rlva, dmglname)
                        vw.makeImport(rlva, "*", dmglname)
                        vw.setComment(rlva, name)

                elif rtype == Elf.R_ARM_GLOB_DAT:
                    symidx = r.getSymTabIndex()
                    sym = elf.getDynSymbol(symidx)
                    ptr = sym.st_value

                    if ptr:
                        logger.info('R_ARM_GLOB_DAT: adding Relocation 0x%x -> 0x%x (%s) ', rlva, ptr, dmglname)
                        if addbase:
                            vw.addRelocation(rlva, vivisect.RTYPE_BASEPTR, ptr)
                        else:
                            vw.addRelocation(rlva, vivisect.RTYPE_BASERELOC, ptr)
                        pname = "ptr_%s" % name

                        if vw.vaByName(pname) is None:
                            vw.makeName(rlva, pname)

                        ptr += baseoff
                        vw.makeImport(ptr, "*", dmglname)
                        vw.setComment(ptr, name)

                    else:
                        logger.info('R_ARM_GLOB_DAT: adding Import 0x%x (%s) ', rlva, dmglname)
                        vw.makeImport(rlva, "*", dmglname)
                        vw.setComment(rlva, name)

                elif rtype == Elf.R_ARM_ABS32:
                    symidx = r.getSymTabIndex()
                    sym = elf.getDynSymbol(symidx)
                    ptr = sym.st_value

                    if ptr:
                        logger.info('R_ARM_ABS32: adding Relocation 0x%x -> 0x%x (%s) ', rlva, ptr, dmglname)
                        vw.addRelocation(rlva, vivisect.RTYPE_BASEPTR, ptr)
                        pname = "ptr_%s" % name
                        if vw.vaByName(pname) is None and len(name):
                            vw.makeName(rlva, pname)

                    else:
                        logger.info('R_ARM_ABS32: adding Import 0x%x (%s) ', rlva, dmglname)
                        vw.makeImport(rlva, "*", dmglname)
                        vw.setComment(rlva, name)

                    vw.setComment(rlva, dmglname)

                elif rtype == Elf.R_ARM_RELATIVE:   # Adjust locations for the rebasing
                    ptr = vw.readMemoryPtr(rlva)
                    logger.info('R_ARM_RELATIVE: adding Relocation 0x%x -> 0x%x (name: %s) ', rlva, ptr, dmglname)
                    vw.addRelocation(rlva, vivisect.RTYPE_BASEPTR, ptr)
                    if len(name):
                        vw.makeName(rlva, dmglname, makeuniq=True)
                        if name != dmglname:
                            vw.setComment(rlva, name)

                elif rtype == Elf.R_ARM_COPY:
                    pass

                else:
                    logger.warning('unknown reloc type: %d %s (at %s)', rtype, name, hex(rlva))
                    logger.info(r.tree())
                    if name:
                        vw.makeName(rlva, dmglname, makeuniq=True)
                        if name != dmglname:
                            vw.setComment(rlva, name)

            if arch in ('a64', 'aarch64'):
                # ARM REL entries require an addend that could be stored as a
                # number or an instruction!
                import envi.archs.arm.const as eaac
                if r.vsHasField('r_addend'):
                    # this is a RELA object, bringing its own addend field!
                    addend = r.r_addend
                else:
                    # otherwise, we have to check the stored value for number or instruction
                    # if it's an instruction, we have to use the immediate value and then
                    # figure out if it's negative based on the instruction!
                    try:
                        temp = vw.readMemoryPtr(rlva)
                        if rtype in Elf.r_armclasses[Elf.R_ARMCLASS_DATA] or rtype in Elf.r_armclasses[Elf.R_ARMCLASS_MISC]:
                            # relocation points to a DATA or MISCELLANEOUS location
                            addend = temp
                        else:
                            # relocation points to a CODE location (ARM, THUMB16, THUMB32)
                            op = vw.parseOpcode(rlva)
                            for oper in op.opers:
                                if hasattr(oper, 'val'):
                                    addend = oper.val
                                    break

                                elif hasattr(oper, 'offset'):
                                    addend = oper.offset
                                    break

                            lastoper = op.opers[-1]
                            if op.mnem.startswith('sub') or \
                                    op.mnem in ('ldr', 'str') and \
                                    hasattr(lastoper, 'pubwl') and \
                                    not (lastoper.pubwl & eaac.PUxWL_ADD):
                                        addend = -addend
                    except Exception as e:
                        logger.exception("ELF: Reloc Addend determination: %s", e)
                        addend = temp

                logger.debug('addend: 0x%x', addend)

                if rtype == Elf.R_AARCH64_JUMP_SLOT:
                    symidx = r.getSymTabIndex()
                    sym = elf.getDynSymbol(symidx)
                    ptr = sym.st_value

                    # quick check to make sure we don't provide this symbol
                    # some toolchains like to point the GOT back at it's PLT entry
                    # that does *not* mean it's not an IMPORT
                    if ptr and not isPLT(vw, ptr):
                        logger.info('R_AARCH64_JUMP_SLOT: adding Relocation 0x%x -> 0x%x (%s) ', rlva, ptr, dmglname)
                        # even if addRelocation fails, still make the name, same thing down in GLOB_DAT
                        if addbase:
                            vw.addRelocation(rlva, vivisect.RTYPE_BASEPTR, ptr)
                        else:
                            vw.addRelocation(rlva, vivisect.RTYPE_BASERELOC, ptr)
                        pname = "ptr_%s_%.8x" % (name, rlva)
                        if vw.vaByName(pname) is None:
                            vw.makeName(rlva, pname)

                        # name the target as well
                        ptr += baseoff
                        # normalize thumb addresses
                        ptr &= -2
                        vw.makeName(ptr, name)
                        vw.setComment(ptr, dmglname)

                    else:
                        logger.info('R_AARCH64_JUMP_SLOT: adding Import 0x%x (%s) ', rlva, dmglname)
                        vw.makeImport(rlva, "*", dmglname)
                        vw.setComment(rlva, name)

                elif rtype == Elf.R_AARCH64_GLOB_DAT:
                    symidx = r.getSymTabIndex()
                    sym = elf.getDynSymbol(symidx)
                    ptr = sym.st_value

                    if ptr:
                        logger.info('R_AARCH64_GLOB_DAT: adding Relocation 0x%x -> 0x%x (%s) ', rlva, ptr, dmglname)
                        if addbase:
                            vw.addRelocation(rlva, vivisect.RTYPE_BASEPTR, ptr)
                        else:
                            vw.addRelocation(rlva, vivisect.RTYPE_BASERELOC, ptr)
                        pname = "ptr_%s" % name

                        if vw.vaByName(pname) is None:
                            vw.makeName(rlva, pname)

                        ptr += baseoff
                        vw.makeImport(ptr, "*", dmglname)
                        vw.setComment(ptr, name)

                    else:
                        logger.info('R_AARCH64_GLOB_DAT: adding Import 0x%x (%s) ', rlva, dmglname)
                        vw.makeImport(rlva, "*", dmglname)
                        vw.setComment(rlva, name)

                elif rtype == Elf.R_AARCH64_ABS64:
                    symidx = r.getSymTabIndex()
                    sym = elf.getDynSymbol(symidx)
                    ptr = sym.st_value

                    if ptr:
                        logger.info('R_AARCH64_ABS64: adding Relocation 0x%x -> 0x%x (%s) ', rlva, ptr, dmglname)
                        vw.addRelocation(rlva, vivisect.RTYPE_BASEPTR, ptr)
                        pname = "ptr_%s" % name
                        if vw.vaByName(pname) is None and len(name):
                            vw.makeName(rlva, pname)

                    else:
                        logger.info('R_AARCH64_ABS64: adding Import 0x%x (%s) ', rlva, dmglname)
                        vw.makeImport(rlva, "*", dmglname)
                        vw.setComment(rlva, name)

                    vw.setComment(rlva, dmglname)

                elif rtype == Elf.R_AARCH64_RELATIVE:   # Adjust locations for the rebasing
                    ptr = vw.readMemoryPtr(rlva)
                    logger.info('R_AARCH64_RELATIVE: adding Relocation 0x%x -> 0x%x (name: %s) ', rlva, ptr, dmglname)
                    vw.addRelocation(rlva, vivisect.RTYPE_BASEPTR, ptr)
                    if len(name):
                        vw.makeName(rlva, dmglname, makeuniq=True)
                        if name != dmglname:
                            vw.setComment(rlva, name)

                elif rtype == Elf.R_AARCH64_COPY:
                    pass

                else:
                    logger.warning('unknown reloc type: %d %s (at %s)', rtype, name, hex(rlva))
                    logger.info(r.tree())
                    if name:
                        vw.makeName(rlva, dmglname, makeuniq=True)
                        if name != dmglname:
                            vw.setComment(rlva, name)


        except vivisect.InvalidLocation as e:
            logger.warning("NOTE\t%r", e)

    return postfix

def isPLT(vw, va):
    '''
    Do a decent check to see if this va is in a PLT section
    '''
    seg = vw.getSegment(va)
    if seg is None:
        return False
    if seg[2].startswith('.plt'):
        return True
    return False

def normName(name):
    '''
    Normalize symbol names.  ie. drop the @@GOBBLEDEGOOK from the end
    '''
    atidx = name.find('@@')
    if atidx > -1:
        name = name[:atidx]
    return name

def demangle(name):
    '''
    Translate C++ mangled name back into the verbose C++ symbol name (with helpful type info)
    '''
    name = normName(name)

    try:
        import cxxfilt
        name = cxxfilt.demangle(name)
    except Exception as e:
        logger.debug('failed to demangle name (%r): %r', name, e)

    return name

def getRelRo(elf):
    status = 0
    for phdr in elf.getPheaders():
        if phdr.p_type == Elf.PT_GNU_RELRO:
            status = 1
            break

    if status:
        for dyn in elf.getDynamics():
            if dyn.d_tag == Elf.DT_FLAGS:
                if dyn.d_value == Elf.DF_BIND_NOW:
                    status = 2

    return ('NONE', 'Partial', 'FULL')[status]

def hasStackCanaries(vw):
    '''
    Check through imports looking for __stack_chk_fail
    '''
    for impva, impsz, imptype, impname in vw.getImports():
        if impname == '*.__stack_chk_fail':
            return True

    return False

def hasNX(elf):
    '''
    Check through ELF Pheaders
    '''
    for phdr in elf.getPheaders():
        if phdr.p_type == Elf.PT_GNU_STACK:
            if phdr.p_flags == 6:
                return True

    return False

def hasPIE(elf):
    '''
    Check through ELF Headers and Dynamics
    '''
    if elf.e_type == Elf.ET_DYN:
        if elf.dyns.get(Elf.DT_DEBUG) is not None:
            return True
        return "DSO"

    return False

def hasRPATH(elf):
    '''
    Check through ELF Dynamics
    '''
    if elf.dyns.get(Elf.DT_RPATH) is not None:
        return True

    return False

def hasRUNPATH(elf):
    '''
    Check through ELF Dynamics
    '''
    if elf.dyns.get(Elf.DT_RUNPATH) is not None:
        return True

    return False

def isStripped(elf):
    '''
    Check through ELF Symbols
    '''
    if len(elf.getSymbols()) == 0:
        return True

    return False


'''
Base Offsets are complicated, so let's break them down a bit.

Base Offsets originally refered to whether to add a base address to values to come up with a virtual address.
Variations of ELFs that regularly cause problems/differences:
    * Shared Objects
    * Kernel Modules / Object files
    * Prelinked SO's and EXEs

Accesses which are impacted by these addresses/offsets:
    * Pointer to Program Headers
    * Pointer to Section Headers
    * Program Headers
    * Sections
    * Dynamics Table
    * Dynamics Entries (see dt_rebase for existing list which require adding base offsets)
    Via either Sections or Dynamics:
        * String Table(s)
        * Symbol Table(s)
        * Relocs Table(s)
        * GOT
        * PLT
        * INIT / FINI
        * INIT_ARRAY / FINI_ARRAY / PREINIT_ARRAY
        * ???  Dwarf?

If a file isn't intended for relocation, this can cause problems because:
    * Pointers can be hard-coded into the instructions themselves
    * ELF headers/Dynamics can not understand the relocatable parts

    
'''



'''
A very simple tool for generating import API definitions from a viv
workspace...
'''

import sys
import argparse

import vivisect


type_lookup = {
    'Unknown': 'int',
    'Pointer': 'void *',
    'FUNCPTR': 'void *',
    'ObjectRef': 'void *',
}

name_lookup = {
    'DWORD': None,
    'Unknown': None,
    'Pointer': 'ptr',
    'ObjectRef': 'obj',
    'FUNCPTR': 'funcptr'
}


def setup():
    ap = argparse.ArgumentParser('Generating impapi definitions from a vivisect workspace')
    ap.add_argument('vw', help='Path to vivisect workspace')
    return ap


def main(argv):

    opts = setup().parse_args(argv)
    vw = vivisect.VivWorkspace()
    vw.loadWorkspace(opts.vw)

    print('# %s' % opts.vw)

    fnames = {}

    for fva, etype, ename, fname in vw.getExports():

        enamekey = ename.lower()
        fnamekey = fname.lower()

        fnames[fname] = True

        # Skip past forwarders
        if not vw.isFunction(fva):
            continue

        rtype, rname, ccname, funcname, args = vw.getFunctionApi(fva)
        args = tuple([(type_lookup.get(t, t), name_lookup.get(t)) for t, name in args])
        print("    '%s.%s':( %r, None, %r, '%s.%s', %r )," % (fnamekey, enamekey, rtype, ccname, fname, ename, args))

    for fwdfname in fnames.keys():

        for rva, name, fwdname in vw.getFileMeta(fwdfname, 'forwarders', ()):
            fwdapi = vw.getImpApi(fwdname)
            if not fwdapi:
                print('    # FIXME unresolved %s -> %s' % (name, fwdname))
                continue

            print("    '%s.%s':%r," % (fwdfname.lower(), name.lower(), fwdapi))


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))




"""
A package of builtin vivisect reports which produce runtime
generated lists of workspace statistics and anomalies.

Each report module must have 2 things. It must have a variable
declared named "columns" which is a list of (colname, coltype)
tuples, and a "report" function which returns a VA Set style
dictionary.
"""

rep_mods = [
    ('Undefined Xrefs/Names', 'vivisect.reports.undeftargets'),
    ('Overlapped Locations', 'vivisect.reports.overlaplocs'),
    ('Function Complexity', 'vivisect.reports.funccomplexity'),
    ('Location Distribution', 'vivisect.reports.locationdist'),
]


def listReportModules():
    '''
    Return a list of (<printname>, <modpath>) tuples for the known
    report types.
    '''
    return list(rep_mods)


def runReportModule(vw, modname):
    mod = vw.loadModule(modname)
    return (mod.columns, mod.report(vw))



'''
A home for the vdb specific memory renderers.
'''
import logging

import envi
import envi.common as e_common
import envi.memcanvas as e_canvas
import envi.memcanvas.renderers as e_canvas_rend

import vivisect.impapi as viv_impapi

import vtrace


logger = logging.getLogger(__name__)

class OpcodeRenderer(e_canvas.MemoryRenderer):

    def __init__(self, trace, arch=envi.ARCH_DEFAULT):
        self.arch = arch
        self.emu_cache = {}  # arch_num: emu instance
        self.pwidth = trace.getPointerSize()
        self.pformat = '0x%%.%dx' % (self.pwidth * 2)

    def _getOpcodePrefix(self, trace, va, op):
        regs = trace.getRegisters()
        regs = {rval: rname for (rname, rval) in regs.items() if rval != 0}

        bp = trace.getBreakpointByAddr(va)
        if bp is not None:
            return ('bp[%d]' % bp.id).ljust(8)

        rname = regs.get(va)
        if rname is not None:
            return rname[:7].ljust(8)

        return '        '

    def _getOpcodeSuffix(self, trace, va, op):
        pc = trace.getProgramCounter()
        if va != pc:
            return ''

        ovals = []
        for o in op.opers:

            if o.isDeref():
                ova = o.getOperAddr(op, trace)
            else:
                ova = o.getOperValue(op, trace)

            sym = None
            if trace.isValidPointer(ova):
                rova = trace.readMemoryFormat(ova, '<P')[0]
                sym = trace.getSymByAddr(rova)

            if sym is None:
                sym = trace.getSymByAddr(ova)

            if sym:
                ovals.append(repr(sym))
            elif o.isDeref():
                ovals.append('[0x%.8x]' % ova)
            else:
                ovals.append('0x%.8x' % ova)

        if [branch for branch, flag in op.getBranches() if flag & envi.BR_COND]:
            emu = self.emu_cache.get(self.arch, vtrace.getEmu(trace))
            emu.setRegisters(trace.getRegisters())
            emu.setProgramCounter(va)
            emu.executeOpcode(op)
            nextpc = emu.getProgramCounter()
            if va + len(op) != nextpc:
                ovals.append('Branch taken: 0x%08x' % nextpc)
            else:
                ovals.append('Branch not taken: 0x%08x' % nextpc)

        return ','.join(ovals)

    def render(self, mcanv, va):
        vastr = self.pformat % va
        # NOTE: we assume the memobj is a trace
        trace = mcanv.mem
        sym = trace.getSymByAddr(va)
        if sym is not None:
            mcanv.addText('\n')
            mcanv.addVaText(str(sym), va=va)
            mcanv.addText(':\n')

        op = trace.parseOpcode(va, arch=self.arch)
        obytes = trace.readMemory(va, op.size)[:8]

        prefix = self._getOpcodePrefix(trace, va, op)
        mcanv.addText(prefix)

        mcanv.addVaText(vastr, va=va)
        mcanv.addText(": %s " % e_common.hexify(obytes).ljust(17))
        op.render(mcanv)

        try:
            suffix = self._getOpcodeSuffix(trace, va, op)
            if suffix:
                mcanv.addText(' ;'+suffix)
        except Exception as e:
            mcanv.addText('; suffix error: %s' % e)

        mcanv.addText("\n")
        return len(op)

class SymbolRenderer(e_canvas.MemoryRenderer):
    def __init__(self, trace):
        a = trace.getMeta("Architecture")
        self.arch = envi.getArchModule(a)
        self.pwidth = self.arch.getPointerSize()

    def render(self, mcanv, va):
        # This is only used with tracer based stuff...
        trace = mcanv.mem
        vastr = self.arch.pointerString(va)
        # NOTE: we assume the memobj is a trace
        trace = mcanv.mem
        p = trace.readMemoryFormat(va, 'P')[0]

        isptr = trace.isValidPointer(p)

        pstr = self.arch.pointerString(p)

        mcanv.addVaText(vastr, va=va)
        mcanv.addText(": ")
        if isptr:
            mcanv.addVaText(pstr, p)
        else:
            mcanv.addText(pstr)

        if isptr:
            sym = trace.getSymByAddr(p, exact=False)
            if sym is not None:
                mcanv.addText(' %s + %d' % (repr(sym), p-int(sym)))

        mcanv.addText('\n')

        return self.pwidth


class DerefRenderer(e_canvas.MemoryRenderer):
    def __init__(self, trace):
        a = trace.getMeta("Architecture")
        self.arch = envi.getArchModule(a)
        self.pwidth = self.arch.getPointerSize()

    def renderData(self, mcanv, va):
        vastr = self.arch.pointerString(va)
        # NOTE: we assume the memobj is a trace
        trace = mcanv.mem
        p = trace.readMemoryFormat(va, 'P')[0]
        isptr = trace.isValidPointer(p)
        pstr = self.arch.pointerString(p)

        vareg = ""
        preg = ""

        regs = trace.getRegisters()
        for name, val in regs.items():
            if val == 0:
                continue
            if val == va:
                vareg = "(%s)" % name
            if val == p:
                preg = "(%s)" % name

        bt = trace.getStackTrace()
        if len(bt) > 1:
            for i in range(1, len(bt)):
                spc, sfc = bt[i]
                if sfc == 0:
                    break
                if spc == 0:
                    break
                if va == spc:
                    vareg = "(savepc)"
                if va == sfc:
                    vareg = "(frame%d)" % i
                if p == spc:
                    preg = "(savepc)"
                if p == sfc:
                    preg = "(frame%d)" % i

        vareg = vareg.ljust(8)
        preg = preg.ljust(8)

        mcanv.addText(" %s: " % str(va - mcanv._canv_beginva).ljust(5))
        mcanv.addVaText(vastr, va=va)
        mcanv.addText(" %s: " % vareg)
        if isptr:
            mcanv.addVaText(pstr, p)
        else:
            mcanv.addText(pstr)

        mcanv.addText(preg)

    def renderMetadata(self, mcanv, va):
        trace = mcanv.mem
        p = trace.readMemoryFormat(va, 'P')[0]
        e_canvas_rend.AutoBytesRenderer().render(mcanv, p)

    def render(self, mcanv, va):
        self.renderData(mcanv, va)
        self.renderMetadata(mcanv, va)
        mcanv.addText('\n')
        return self.arch.getPointerSize()


class StackRenderer(DerefRenderer):
    def __init__(self, trace):
        DerefRenderer.__init__(self, trace)

    def render(self, mcanv, va):
        trace = mcanv.mem
        if va != trace.getStackCounter():
            return DerefRenderer.render(self, mcanv, va)

        pc = trace.getProgramCounter()
        sym, is_thunk = trace.getSymByAddrThunkAware(pc)
        if sym is None:
            return DerefRenderer.render(self, mcanv, va)

        # TODO: this code also exists in win32stealth and in hookbreakpoint
        # we should put this somewhere common
        platform = trace.getMeta('Platform')
        arch = trace.getMeta('Architecture')
        impapi = viv_impapi.getImportApi(platform, arch)
        cc_name = impapi.getImpApiCallConv(sym)
        emu = vtrace.getEmu(trace)
        cc = emu.getCallingConvention(cc_name)
        args_def = impapi.getImpApiArgs(sym)
        if args_def is None:
            # sym did not exist in impapi :(
            logger.warning('sym but no impapi match: {}'.format(sym))
            return DerefRenderer.render(self, mcanv, va)

        argc = len(args_def)

        curop = trace.parseOpcode(trace.getProgramCounter())

        # use the calling convention to retrieve the args
        args = None
        if curop.isCall() or is_thunk:
            args = cc.getPreCallArgs(trace, argc)
        else:
            args = cc.getCallArgs(trace, argc)

        # since we are 'normalizing' the calls by visualizing all calling
        # conventions in a stdcall fashion, some args (like the ones in
        # registers don't have a stack va.
        mcanv.addText('%s :\n' % sym)
        fmt = '  arg%%d (%%s) 0x%%0%dx %%s\n' % (trace.getPointerSize()*2,)
        for index, arg in enumerate(args):
            argtype = args_def[index][0]
            argva = arg
            if trace.isValidPointer(arg):
                argva = trace.readMemoryFormat(arg, 'P')[0]
            smc = e_canvas.StringMemoryCanvas(trace)
            e_canvas_rend.AutoBytesRenderer(maxrend=64).render(smc, argva)
            desc = str(smc)
            mcanv.addText(fmt % (index, argtype, arg, desc))
        mcanv.addText('-' * 5)
        mcanv.addText('\n')

        return DerefRenderer.render(self, mcanv, va)



import logging

from PyQt5.QtWidgets import *

logger = logging.getLogger(__name__)

class FieldAdder:

    def __init__(self, splitchar='.'):
        self.kids = {}
        self.splitchar = splitchar
        self._dyn_callback = None

    def _addMenuFields(self, plist):
        parent = self
        kid = self
        for p in plist:
            kid = parent.kids.get(p)
            if kid is None:
                kid = VQMenu(p, parent=parent, splitchar=self.splitchar)
                action = parent.addMenu(kid)
                parent.kids[p] = kid
            parent = kid

        return kid

    def _addDynActions(self):
        self.clear()
        for name in self._dyn_callback():
            act = QAction(name, self)
            cb = ActionCall(self._dyn_callback, name)
            act.triggered.connect(cb)
            self.addAction(act)

    def addDynMenu(self, pathstr, callback):
        '''
        Add a dynamic menu item which will populate on-hover by asking
        the specified callback for a list of actions.

        Example:
            def callback(name=None):

                if name is None:
                    return ('one', 'two', 'three')

                print('SELECTED: %s' % name)

            m.addDynMenu( callback )
        '''
        plist = pathstr.split(self.splitchar)
        menu = self._addMenuFields( plist )
        menu._dyn_callback = callback
        menu.aboutToShow.connect(menu._addDynActions)

    def addField(self, pathstr, callback=None, args=(), tip=None):
        plist = pathstr.split(self.splitchar)
        kid = self._addMenuFields(plist[:-1])

        acall = ActionCall( callback, *args )
        action = QAction(plist[-1], kid)
        action.triggered.connect(acall)

        if tip: action.setStatusTip(tip)

        kid.addAction(action)

        return kid

class VQMenuBar(FieldAdder, QMenuBar):
    def __init__(self, parent=None, splitchar='.'):
        QMenuBar.__init__(self, parent=parent)
        FieldAdder.__init__(self, splitchar=splitchar)

class VQMenu(FieldAdder, QMenu):

    def __init__(self, name, parent=None, splitchar='.'):
        QMenu.__init__(self, name, parent=parent)
        FieldAdder.__init__(self, splitchar=splitchar)

class ActionCall:

    def __init__(self, callback, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self.callback = callback

    def __call__(self):
        try:
            retval = self.callback(*self.args, **self.kwargs)
            return retval
        except Exception as e:
            logger.exception("ActionCall failed on %s with error: %s ", repr(self.callback), str(e))



import os
import sys
import logging
import zipfile
import binascii
import optparse
import tempfile

import cobra
import cobra.remoteapp as c_remoteapp
import envi.common as e_common

logger = logging.getLogger(__name__)


def release():

    parser = optparse.OptionParser()

    parser.add_option('--cacert', dest='cacert', default=None)
    parser.add_option('--sslkey', dest='sslkey', default=None)
    parser.add_option('--sslcert', dest='sslcert', default=None)

    opts, argv = parser.parse_args()
    pyzfile, appuri = argv

    with open(__file__, 'rb') as f:
        mainsrc = f.read()
    mainlines = mainsrc.split('\n')[:-2]

    castr = 'None'
    keystr = 'None'
    certstr = 'None'

    if opts.cacert:
        with open(opts.cacert, 'rb') as f:
            castr = '"%s"' % e_common.hexify(f.read())

    if opts.sslkey:
        with open(opts.sslkey, 'rb') as f:
            keystr = '"%s"' % e_common.hexify(f.read())

    if opts.sslcert:
        with open(opts.sslcert, 'rb') as f:
            certstr = '"%s"' % e_common.hexify(f.read())

    mainlines.append('    appuri="%s"' % appuri)
    mainlines.append('    cacrt=%s' % castr)
    mainlines.append('    sslkey=%s' % keystr)
    mainlines.append('    sslcert=%s' % certstr)
    mainlines.append('    main(appuri, cacrt=cacrt, sslcert=sslcert, sslkey=sslkey)')

    mainsrc = '\n'.join(mainlines)

    pyz = zipfile.PyZipFile(pyzfile, 'w')
    pyz.writepy('cobra')
    pyz.writestr('__main__.py', mainsrc)
    pyz.close()


def dumpfile(hexbytes, filepath):
    with open(filepath, 'wb') as f:
        f.write(binascii.unhexlify(hexbytes))
    return filepath


def main(uri, cacrt=None, sslcert=None, sslkey=None):
    e_common.initLogging(logger, 'INFO')
    if any([cacrt, sslcert, sslkey]):
        scheme, host, port, name, urlparams = cobra.chopCobraUri(uri)
        builder = cobra.initSocketBuilder(host, port)

        tempdir = tempfile.mkdtemp()
        if cacrt:
            cafile = dumpfile(cacrt, os.path.join(tempdir, 'ca.crt'))
            builder.setSslCa(cafile)

        if sslkey:
            keyfile = dumpfile(sslkey, os.path.join(tempdir, 'client.key'))
            certfile = dumpfile(sslcert, os.path.join(tempdir, 'client.crt'))
            builder.setSslClientCert(certfile, keyfile)

    try:
        c_remoteapp.getAndRunApp(uri)
    except Exception as e:
        logger.warning('error: %s', e)


if __name__ == '__main__':
    sys.exit(release())  # this *must* be the *last* line...



'''
A package for each of the different layout managers.
'''


# Some helper utils...
def exit_pos(ninfo):
    x,y = ninfo.get('position')
    xsize, ysize = ninfo.get('size', (0,0))
    return (x + xsize/2, y + ysize)

def entry_pos(ninfo):
    x,y = ninfo.get('position')
    xsize, ysize = ninfo.get('size', (0,0))
    return (x + xsize/2, y)

def center_pos(ninfo):
    x,y = ninfo.get('position')
    xsize, ysize = ninfo.get('size', (0,0))
    return (x + (xsize/2), y + (ysize/2))

class GraphLayout:

    '''
    A graph layout uses several graph meta properties and node properties
    to communicate with a renderer which is expected to display the graph:

    size = ( width, height )    - Set by the renderer
    position = ( x, y )         - Set by the layout
    repr = <display text>       - A fallback for what to display on a node
    '''

    def __init__(self, graph):
        self.graph = graph

    def layoutGraph(self):
        '''
        Layout the graph nodes and edges
        '''
        raise Exception('%s must implement layoutGraph()!' % self.__class__.__name__)

    def getLayoutSize(self):
        raise Exception('%s must implement getLayoutSize()!' % self.__class__.__name__)

    def renderGraph(self, rend):
        '''
        Render the graph to the given renderer.
        '''
        rend.setNodeSizes(self.graph)
        self.layoutGraph()
        width, height = self.getLayoutSize()
        self.graph.setMeta('size', (width, height) )

        rend.renderGraph()



import os

import envi.common as e_common
import vivisect.symboliks.expression as v_s_expr

from vivisect.const import *
from vivisect.symboliks.common import *

class SymbolikEmulator:

    # FIXME possible  _sym_width set automatically from vw?
    # FIXME possible "sym factory" mixin that builds Vars/Consts/Opers etc?
    # ( v = emu.initSymvar('woot')  ( with size default etc.. )

    __width__ = None    # Extenders *must* set this.

    def __init__(self, vw):
        '''
        The SymbolikEmulator is used to keep state on symbolik execution.
        '''
        self._sym_meta = {}
        self._sym_vars = {}
        self._sym_mem = {}
        self._sym_rseed = ''
        self._sym_vw = vw
        self._sym_cconvs = {}    # function emulation can be calling convention aware

        self._sym_expr_parser = v_s_expr.SymbolikExpressionParser(defwidth=vw.psize)

    def getSymSnapshot(self):
        return (dict(self._sym_meta),
                dict(self._sym_vars),
                dict(self._sym_mem),
                self._sym_rseed)

    def setSymSnapshot(self, snap):
        (self._sym_meta,
         self._sym_vars,
         self._sym_mem,
         self._sym_rseed) = snap

    def setMeta(self, name, val):
        '''
        Store metadata in the emulator instance for later.
        '''
        self._sym_meta[name] = val

    def getMeta(self, name, default=None):
        '''
        Retrieve previously stored emulator metadata.
        '''
        return self._sym_meta.get(name, default)

    def parseExpression(self, expr, update=True):
        sym = self._sym_expr_parser.parseExpression(expr)
        if update:
            sym = sym.update(emu=self)
        return sym

    def solveExpression(self, expr, update=True):
        sym = self._sym_expr_parser.parseExpression(expr)
        if update:
            sym = sym.update(emu=self)
        return sym.solve(emu=self)

    def applyEffects(self, effects):
        '''
        Apply the given effects to the emulator.  Return a list of updated
        effects which reflect the state during emulation.
        '''
        return [e.applyEffect(self) for e in effects]

    def applyFunctionCall(self, funcsym):
        '''
        This API allows arch/platform specific emulators to handle (and expand
        out) function calls to modify the emulator state.  A symbolik emulator
        by itself is *not* function aware, so this does nothing, but on
        SymbolikFunctionEmulator instances this will modify the machine state
        to account for the function call...
        '''
        pass

    def addRandomSeed(self):
        '''
        Add a random seed to the name hashing subsystem.  This will
        produce values which may *only* be compared to other values
        generated by this solver.  It also lets you calculate deltas
        by solving for deltas between two symbols in different seed
        inputs to see if they are likely arithmetically related.
        '''
        self._sym_rseed = e_common.hexify(os.urandom(10))

    def getRandomSeed(self):
        return self._sym_rseed

    def readSymMemory(self, symaddr, symsize, vals=None):
        '''
        The readSymMemory API is designed to read from the given
        symbolik address for the specified symbolik length.  If the
        current symbolik emulator has no knowledge of the state of
        the given memory symbol, None is returned.
        '''
        addrval = symaddr.solve(emu=self, vals=vals)

        # check for a previous write first...
        symmem = self._sym_mem.get(addrval)
        if symmem is not None:
            symaddr, symval = symmem
            return symval

        # If we have a workspace, check it for meaningful
        # symbols etc...
        if self._sym_vw is not None:
            # Make a special check for imports...
            loc = self._sym_vw.getLocation(addrval)
            if loc is not None:
                lva, lsize, ltype, linfo = loc
                if ltype == LOC_IMPORT:
                    return Var(linfo, self.__width__)

        return None

    def writeSymMemory(self, symaddr, symval, vals=None):
        # FIXME handle memory offsets etc...
        # FIXME handle write size.. (using isDiscrete?)
        addrval = symaddr.solve(emu=self, vals=vals)
        # sizeval = symsize.solve(slvctx=self)
        self._sym_mem[addrval] = (symaddr, symval)

    def setSymVariable(self, name, symval, width=None):
        if width is None:
            width = self.__width__
        self._sym_vars[name] = symval

    def getSymVariable(self, name, create=True):
        '''
        Get the current state of a named symbolic variable.

        Example:
            v = self.getSymVariable('eax')
        '''
        ret = self._sym_vars.get(name)
        if ret is None and create:
            return Var(name, self.__width__)

        return ret

    def getSymVariables(self):
        '''
        Retrieve a list of (<varname>, <varsym>) tuples.

        Example:
            for vname, vsym in t.getSymVariables():
                print('%s = %s' % (vname, str(vsym)))
        '''
        return list(self._sym_vars.items())



import sys
import shlex

import vtrace


class ShellcodeLoadNotifier(vtrace.Notifier):
    def __init__(self, db, fname):
        self.db = db
        self.fname = fname

    def notify(self, event, trace):
        if event != vtrace.NOTIFY_BREAK:
            return

        bytez = None
        with open(self.fname, 'rb') as f:
            bytez = f.read()

        va = trace.allocateMemory(len(bytez))
        trace.writeMemory(va, bytez)

        self.db.vprint('Loaded shellcode at address: 0x%x' % va)

        trace.setProgramCounter(va)
        trace.addBreakByAddr(va)

        self.db.deregisterNotifier(vtrace.NOTIFY_BREAK, self)


def execsc(db, line):
    '''
    Load and execute shellcode stub from file

    Usage: execsc <fname>
    '''
    argv = shlex.split(line)
    if len(argv) != 1:
        return db.do_help('execsc')

    cmd = sys.executable
    trace = db.newTrace()
    db.vprint('Executing %s' % cmd)

    sc = ShellcodeLoadNotifier(db, argv[0])
    db.registerNotifier(vtrace.NOTIFY_BREAK, sc)

    trace.execute(cmd)


def vdbExtension(vdb, trace):
    vdb.registerCmdExtension(execsc)



"""
PPC Support Module (not done)
"""
# Copyright (C) 2007 Invisigoth - See LICENSE file for details
class PpcMixin:
    def archAddWatchpoint(self, address):
        pass

    def archRemWatchpoint(self, address):
        pass

    def archCheckWatchpoint(self, address):
        pass

    def getStackTrace(self):
        self.requireAttached()
        return []

    def getBreakInstruction(self):
        # twi 0x14, r0, 0 
        # trap if r0 is (>=:unsigned) 0
        return "\x0e\x80\x00\x00"

    def archGetPcName(self):
        return "r0"

    def archGetSpName(self):
        return "r1"

    def platformCall(self, address, args, convention=None):
        pass



import unittest

from vivisect.symboliks.common import *
from vivisect.symboliks.expression import symexp

class TestReduceCase(unittest.TestCase):
    '''
    tests basic algebraic reductions
    '''
    def assertReduce(self, s1, s2):
        sym1 = symexp(s1).reduce(foo=True)
        sym2 = symexp(s2)
        self.assertEqual(str(sym1), str(sym2))

    def assertNotReduce(self, s1, s2):
        sym1 = symexp(s1).reduce()
        sym2 = symexp(s2)
        self.assertNotEqual(str(sym1), str(sym2))

    def test_symboliks_reduce_and_consts(self):
        self.assertReduce('255 & 255', '255')
        self.assertReduce('0xff & 0xf0', '0xf0')

    def test_symboliks_reduce_op_and(self):
        self.assertReduce('(foo & 0xff) & 0xff', 'foo & 0xff')
        self.assertReduce('(0xff & foo) & 0xff', 'foo & 0xff')
        self.assertReduce('0xff & (foo & 0xff)', 'foo & 0xff')
        self.assertReduce('0xff & (0xff & foo)', 'foo & 0xff')

        self.assertReduce('foo & 0', '0')
        self.assertReduce('0 & foo', '0')

        # reduce & with 0 regardless of width
        self.assertReduce('foo[5] & 0', '0')
        # we only support slices on memory
        # self.assertReduce('foo[2:4] & 0', '0')
        self.assertReduce('mem[2:4] & 0', '0')
        self.assertReduce('0[1] & foo', '0')

        # reduce & umax of foo to just foo
        self.assertReduce('foo[1] & 0xff', 'foo')
        self.assertReduce('foo[2] & 0xffff', 'foo')
        self.assertReduce('foo[4] & 0xffffffff', 'foo')

        self.assertNotReduce('foo[4] & 0xff', 'foo')
        self.assertNotReduce('foo[4] & 0xffff', 'foo')

    def test_symboliks_reduce_op_xor(self):
        self.assertReduce('foo ^ foo', '0')
        self.assertReduce('foo ^ 0', 'foo')

    def test_symboliks_reduce_op_mul(self):
        self.assertReduce('foo * 0', '0')
        self.assertReduce('foo * 1', 'foo')
        self.assertReduce('(foo * 10) / 2', 'foo * 5')
        self.assertReduce('(foo * 10) * 2', 'foo * 20')

    @unittest.skip('Need to implement more reducers')
    def test_symboliks_reduce_op_divmul(self):
        self.assertReduce('(foo / 2) * 2', 'foo')
        self.assertReduce('(foo / 2) * 10', 'foo * 5')

    def test_symboliks_reduce_op_div(self):
        self.assertReduce('0 / foo', '0')
        self.assertReduce('foo / foo', '1')

    def test_symboliks_reduce_op_rshift(self):
        self.assertReduce('0 >> foo', '0')
        self.assertReduce('foo >> 0', 'foo')
        self.assertReduce('3 >> 1', '1')
        self.assertReduce('5347 >> 6', '83')

    def test_symboliks_reduce_op_lshift(self):
        self.assertReduce('0 << foo', '0')
        self.assertReduce('foo << 0', 'foo')
        self.assertReduce('2 << 2', '8')
        self.assertReduce('7 << 2', '28')
        self.assertReduce('1 << 2', '4')

    def test_symboliks_reduce_op_pow(self):
        self.assertReduce('0 ** foo', '0')
        self.assertReduce('1 ** foo', '1')
        self.assertReduce('foo ** 0', '1')

    def test_symboliks_reduce_op_addsub(self):
        self.assertReduce('(foo + 255) + 10', 'foo + 265')
        self.assertReduce('(255 + foo) + 10', 'foo + 265')
        self.assertReduce('(255 + foo) + 10', 'foo + 265')
        self.assertReduce('(foo - 10) + 255', 'foo + 245')
        self.assertReduce('(255 - foo) + 10', '(0 - foo) + 265')
        self.assertReduce('(foo + 255) - 10', 'foo + 245')
        self.assertReduce('(foo - 255) - 10', 'foo - 265')
        self.assertReduce('255 - (foo - 10)', '265 - foo')
        self.assertReduce('255 - (foo + 10)', '245 - foo')
        self.assertReduce('255 - (10 - foo)', 'foo + 245')
        self.assertReduce('(foo + 255) + (bar + 10)', '(foo + bar) + 265')
        self.assertReduce('(foo - 10) + (bar + 255)', '(bar + foo) + 245')
        self.assertReduce('(foo - 255) + (bar - 10)', '(foo + bar) - 265')
        self.assertReduce('20 - (10 - eax)', 'eax + 10')
        self.assertReduce('(foo + 255) - (bar + 10)', '(foo - bar) + 245')
        self.assertReduce('(foo + 255) - (10 + bar)', '(foo - bar) + 245')
        self.assertReduce('(foo + 255) + (bar - 10)', '(foo + bar) + 245')
        self.assertReduce('(foo + 255) + (10 - bar)', '(foo - bar) + 265')
        self.assertReduce('(foo - 255) - (bar + 10)', '(foo - bar) - 265')
        self.assertReduce('(255 - foo) - (bar + 10)', '(( 0 - foo ) - bar) + 245')
        self.assertReduce('(foo + 255) - (bar - 10)', '(foo - bar) + 265')
        self.assertReduce('(foo - 255) - (bar - 275)', '(foo - bar) + 20')
        self.assertReduce('foo + 0', 'foo')
        self.assertReduce('0 + foo', 'foo')
        self.assertReduce('0 + 0', '0')
        self.assertReduce('2 + 2', '4')
        self.assertReduce('2 - 2', '0')
        self.assertReduce('foo + (0 - bar)', 'foo - bar')
        self.assertReduce('foo + (0 + bar)', 'foo + bar')
        self.assertReduce('foo - (0 + bar)', 'foo - bar')
        # TODO: More generic ast reducer
        # self.assertReduce('(foo + bar) - (bar + baz)', 'foo - baz')
        self.assertReduce('0 - (0 - foo)', 'foo')
        self.assertReduce('0 + (0 + foo)', 'foo')
        self.assertReduce('0 - (0 + foo)', '0 - foo')

    def test_symboliks_reduce_varsub(self):
        esp = Var('esp', width=4)
        for i in range(100):
            esp -= Const(4, width=4)
        self.assertEqual(str(esp.reduce()), '(esp - 400)')

    def test_symboliks_reduce_funcargs_multipass(self):
        op = (Const(0x1000, 8) - Const(0xb90, 8)) - Const(0x60, 8)
        arg = (Mem(Const(0x14000, 8), Const(8, 8)) ^ op) ^ op
        expr = Call(Const(0x400, 8), Const(8, 8), argsyms=[arg,])

        expr = expr.reduce()
        self.assertEqual(str(expr), '1024(mem[0x00014000:8])')

        op = (Const(0x1000, 8) - Const(0xb90, 8)) - Const(0x60, 8)
        arg = (Mem(Const(0x14000, 8), Const(8, 8)) ^ op) ^ op
        expr = Call(Const(0x400, 8), Const(8, 8), argsyms=[arg,])
        expr = expr.reduce(foo=True)
        self.assertEqual(str(expr), '1024(mem[0x00014000:8])')
